from __future__ import absolute_import
import random
from unittest.case import TestCase

from omegaml import Omega
from omegaml.mdataframe import MDataFrame
from omegaml.store import qops
from omegaml.store.query import Filter
from omegaml.store.queryops import GeoJSON
import pandas as pd
from six.moves import range
from pandas.util.testing import assert_frame_equal
# see https://gist.github.com/miraculixx/f01304186fc47d041da5a712774ac487
locations = [{'location': {'coordinates': [-74.0059413, 40.7127837],
                           'type': 'Point'},
              'place': 'New York'},
             {'location': {'coordinates': [6.1431577, 46.2043907],
                           'type': 'Point'},
              'place': 'Geneva'},
             {'location': {'coordinates': [7.4474468, 46.9479739],
                           'type': 'Point'},
              'place': 'Bern'},
             {'location': {'coordinates': [8.541694, 47.3768866],
                           'type': 'Point'},
              'place': 'Zurich'}]


class FilterQueryTests(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        df = self.df = pd.DataFrame({'x': list(range(0, 10)) + list(range(0, 10)),
                                     'y': random.sample(list(range(0, 100)), 20)})
        om = self.om = Omega()
        om.datasets.put(df, 'sample', append=False)
        self.coll = om.datasets.collection('sample')

    def tearDown(self):
        TestCase.tearDown(self)

    def test_filter(self):
        coll = self.coll
        df = self.df
        result = Filter(coll, x=0).value
        testdf = df[df.x == 0]
        self.assertTrue(result.equals(testdf))

    def test_filter_in(self):
        coll = self.coll
        df = self.df
        result = Filter(coll, x__in=[1, 2, 3]).value
        testdf = df[df.x.isin([1, 2, 3])]
        self.assertTrue(result.equals(testdf))

    def test_filter_and(self):
        coll = self.coll
        df = self.df
        y = int(df.y.unique()[0])
        result = Filter(coll, x=0, y=y).value
        testdf = df[(df.x == 0) & (df.y == y)]
        self.assertTrue(result.equals(testdf))

    def test_filter_or(self):
        coll = self.coll
        df = self.df
        result = Filter(coll, x=0, y__gt=5).value
        testdf = df[(df.x == 0) & (df.y > 5)]
        self.assertTrue(result.equals(testdf))

    def test_filter_near(self):
        om = self.om
        # create a dataframe with geo locations
        geodf = pd.DataFrame(locations)
        geodf['location'] = geodf.location.apply(lambda v: GeoJSON(v))
        om.datasets.put(geodf, 'geosample', append=False, index='@location')
        coll = om.datasets.collection('geosample')
        # closest place
        result = Filter(coll,
                        location__near=dict(location=(8.541694, 47.3768866), maxd=1))
        places = result.value.place.unique()
        self.assertEqual(places, ['Zurich'])
        # ordered by distance
        result = Filter(coll,
                        location__near=dict(location=(8.541694, 47.3768866)))
        places = list(result.value.place.unique())
        self.assertListEqual(places, 'Zurich,Bern,Geneva,New York'.split(','))
        # use tuple (lon, lat, maxd)
        result = Filter(coll,
                        location__near=(8.541694, 47.3768866, 1))
        places = list(result.value.place.unique())
        self.assertListEqual(places, 'Zurich'.split(','))
        # use tuple (lon, lat, mind, maxd)
        result = Filter(coll,
                        location__near=(8.541694, 47.3768866, 0, 100))
        places = list(result.value.place.unique())
        self.assertListEqual(places, 'Zurich'.split(','))

    def test_filter_subdoc(self):
        coll = self.coll
        coll.update_many(qops.IS(x=qops.LT(5)), qops.SET('subdoc.a', 99))
        coll.update_many(qops.IS(x=qops.GTE(5)), qops.SET('subdoc.a', 0))
        result = Filter(coll, subdoc__a__lt=10).value
        self.assertEqual(set(result.x.unique()), set(range(5, 10)))

    def test_query_null(self):
        om = self.om
        df = pd.DataFrame({'x': list(range(0, 5)),
                           'y': [1, 2, 3, None, None]})
        om.datasets.put(df, 'foox', append=False)
        result = om.datasets.get('foox', y__isnull=True)
        test = df[df.isnull().any(axis=1)]
        assert_frame_equal(result, test)



import pandas as pd
import numpy as np

from omegaml.util import ensure_json_serializable


class GenericModelResource(object):
    def __init__(self, om):
        self.om = om

    def put(self, model_id, query, payload):
        """
        Args:
            model_id (str): the name of the model
            query (dict): the query parameters
            payload (dict): the json body

        Returns:
            dict(model: id, result: dict)
        """
        data = (payload or {}).get('data')
        dataset = query.get('datax')
        if data is not None:
            columns = payload.get('columns')
            shape = payload.get('shape')
            df = pd.DataFrame(data)[columns]
            if shape is not None:
                assert len(columns) == 1, "only 1 column allowed to be reshaped"
                col = columns[0]
                df[col] = df[col].apply(lambda v: np.array(v).reshape(shape))
                df = np.stack(df[col])
            promise = self.om.runtime.model(model_id).predict(df)
        elif dataset:
            promise = self.om.runtime.model(model_id).predict(dataset)
        else:
            raise ValueError('require either "data" key in body, or ?datax=dataset')
        result = promise.get()
        # if we have a single column, get as a list
        # -- refactored from EE
        if hasattr(result, 'shape'):
            if len(result.shape) > 1 and result.shape[1] == 1:
                result = result[:, 0]
            result = result.tolist()
        return {'model': model_id, 'result': ensure_json_serializable(result)}



from __future__ import absolute_import

import re
from uuid import uuid4

import datetime
import gridfs
import yaml
from croniter import croniter
from mongoengine.fields import GridFSProxy
from nbconvert.preprocessors.execute import ExecutePreprocessor
from nbformat import read as nbread, write as nbwrite, v4 as nbv4
from six import StringIO, BytesIO

from omegaml.documents import MDREGISTRY
from omegaml.notebook.jobschedule import JobSchedule
from omegaml.store import OmegaStore
from omegaml.util import settings as omega_settings


class OmegaJobs(object):
    """
    Omega Jobs API
    """

    # TODO this class should be a proper backend class with a mixin for ipynb

    _nb_config_magic = 'omega-ml', 'schedule', 'run-at', 'cron'
    _dir_placeholder = '_placeholder.ipynb'

    def __init__(self, prefix=None, store=None, defaults=None):
        self.defaults = defaults or omega_settings()
        prefix = prefix or 'jobs'
        self.store = store or OmegaStore(prefix=prefix)
        self.kind = MDREGISTRY.OMEGAML_JOBS
        self._include_dir_placeholder = True
        # convenience so you can do om.jobs.schedule(..., run_at=om.jobs.Schedule(....))
        self.Schedule = JobSchedule

    def __repr__(self):
        return 'OmegaJobs(store={})'.format(self.store.__repr__())

    @property
    def _db(self):
        return self.store.mongodb

    @property
    def _fs(self):
        return self.store.fs

    def collection(self, name):
        if not name.endswith('.ipynb'):
            name += '.ipynb'
        return self.store.collection(name)

    def drop(self, name, force=False):
        meta = self.metadata(name)
        name = meta.name  # ensure we get the actual name
        return self.store.drop(name, force=force)

    def metadata(self, name):
        meta = self.store.metadata(name)
        if meta is None and not name.endswith('.ipynb'):
            name += '.ipynb'
            meta = self.store.metadata(name)
        return meta

    def exists(self, name):
        return len(self.store.list(name)) + len(self.store.list(name + '.ipynb')) > 0

    def put(self, obj, name, attributes=None):
        """
        Store a NotebookNode

        :param obj: the NotebookNode to store
        :param name: the name of the notebook
        """
        if not name.endswith('.ipynb'):
            name += '.ipynb'
        sbuf = StringIO()
        bbuf = BytesIO()
        # nbwrite expects string, fs.put expects bytes
        nbwrite(obj, sbuf, version=4)
        sbuf.seek(0)
        bbuf.write(sbuf.getvalue().encode('utf8'))
        bbuf.seek(0)
        # see if we have a file already, if so replace the gridfile
        meta = self.store.metadata(name)
        if not meta:
            filename = uuid4().hex
            fileid = self._fs.put(bbuf, filename=filename)
            meta = self.store._make_metadata(name=name,
                                             prefix=self.store.prefix,
                                             bucket=self.store.bucket,
                                             kind=self.kind,
                                             attributes=attributes,
                                             gridfile=GridFSProxy(grid_id=fileid))
            meta = meta.save()
        else:
            filename = uuid4().hex
            meta.gridfile.delete()
            fileid = self._fs.put(bbuf, filename=filename)
            meta.gridfile = GridFSProxy(grid_id=fileid)
            meta = meta.save()
        # set config
        nb_config = self.get_notebook_config(name)
        meta_config = meta.attributes.get('config', {})
        if nb_config:
            meta_config.update(dict(**nb_config))
            meta.attributes['config'] = meta_config
        meta = meta.save()
        if not name.startswith('results') and ('run-at' in meta_config):
            meta = self.schedule(name)
        return meta

    def get(self, name):
        """
        Retrieve a notebook and return a NotebookNode
        """
        if not name.endswith('.ipynb'):
            name += '.ipynb'
        meta = self.store.metadata(name)
        if meta:
            try:
                outf = meta.gridfile
            except gridfs.errors.NoFile as e:
                raise e
            # nbwrite wants a string, outf is bytes
            sbuf = StringIO()
            data = outf.read()
            if data is None:
                msg = 'Expected content in {name}, got None'.format(**locals())
                raise ValueError(msg)
            sbuf.write(data.decode('utf8'))
            sbuf.seek(0)
            nb = nbread(sbuf, as_version=4)
            return nb
        else:
            raise gridfs.errors.NoFile(
                ">{0}< does not exist in jobs bucket '{1}'".format(
                    name, self.store.bucket))

    def create(self, code, name):
        """
        create a notebook from code

        :param code: the code as a string
        :param name: the name of the job to create
        :return: the metadata object created
        """
        cells = []
        cells.append(nbv4.new_code_cell(source=code))
        notebook = nbv4.new_notebook(cells=cells)
        # put the notebook
        meta = self.put(notebook, name)
        return meta

    def get_fs(self, collection=None):
        # legacy support
        return self._fs

    def get_collection(self, collection):
        """
        returns the collection object
        """
        # FIXME this should use store.collection
        return getattr(self.store.mongodb, collection)

    def list(self, jobfilter='.*', raw=False, **kwargs):
        """
        list all jobs matching filter.
        filter is a regex on the name of the ipynb entry.
        The default is all, i.e. `.*`
        """
        job_list = self.store.list(regexp=jobfilter, raw=raw, **kwargs)
        name = lambda v: v.name if raw else v
        if not self._include_dir_placeholder:
            job_list = [v for v in job_list if not name(v).endswith(self._dir_placeholder)]
        return job_list

    def get_notebook_config(self, nb_filename):
        """
        returns the omegaml script config on
        the notebook's first cell

        If there is no config cell or the config cell is invalid raises
        a ValueError
        """
        notebook = self.get(nb_filename)
        config_cell = None
        config_magic = ['# {}'.format(kw) for kw in self._nb_config_magic]
        for cell in notebook.get('cells'):
            if any(cell.source.startswith(kw) for kw in config_magic):
                config_cell = cell
        if not config_cell:
            return {}
        yaml_conf = '\n'.join(
            [re.sub('#', '', x, 1) for x in str(
                config_cell.source).splitlines()])
        try:
            yaml_conf = yaml.safe_load(yaml_conf)
            config = yaml_conf.get(self._nb_config_magic[0], yaml_conf)
        except Exception:
            raise ValueError(
                'Notebook configuration cannot be parsed')
        # translate config to canonical form
        # TODO refactor to seperate method / mapped translation functions
        if 'schedule' in config:
            config['run-at'] = JobSchedule(text=config.get('schedule', '')).cron
        if 'cron' in config:
            config['run-at'] = JobSchedule.from_cron(config.get('cron')).cron
        if 'run-at' in config:
            config['run-at'] = JobSchedule.from_cron(config.get('run-at')).cron
        return config

    def run(self, name):
        """
        Run a job immediately

        The job is run and the results are stored in the given filename

        :param name: the name of the jobfile
        :return: the metadata of the job
        """
        return self.run_notebook(name)

    def run_notebook(self, name, event=None):
        """
        run a given notebook immediately.
        the job parameter is the name of the job script as in ipynb.
        Inserts and returns the Metadata document for the job.
        """
        notebook = self.get(name)
        meta_job = self.metadata(name)
        ts = datetime.datetime.now()
        # execute
        ep_kwargs = {
            'timeout': None
        }
        ep_kwargs.update(meta_job.kind_meta.get('ep_kwargs', {}))
        try:
            ep = ExecutePreprocessor(**ep_kwargs)
            ep.preprocess(notebook, {'metadata': {'path': '/'}})
        except Exception as e:
            status = 'ERROR'
            message = str(e)
        else:
            status = 'OK'
            message = ''
            # record results
            meta_results = self.put(
                notebook, 'results/{name}_{ts}'.format(**locals()))
            meta_results.attributes['source_job'] = name
            meta_results.save()
            job_results = meta_job.attributes.get('job_results', [])
            job_results.append(meta_results.name)
            meta_job.attributes['job_results'] = job_results
        # record final job status
        job_runs = meta_job.attributes.get('job_runs', [])
        runstate = {
            'status': status,
            'ts': ts,
            'message': message,
            'results': meta_results.name if status == 'OK' else None
        }
        job_runs.append(runstate)
        meta_job.attributes['job_runs'] = job_runs
        # set event run state if event was specified
        if event:
            attrs = meta_job.attributes
            triggers = attrs['triggers'] = attrs.get('triggers', [])
            scheduled = (trigger for trigger in triggers
                         if trigger['event-kind'] == 'scheduled')
            for trigger in scheduled:
                if event == trigger['event']:
                    trigger['status'] = status
                    trigger['ts'] = ts
        return meta_job.save()

    def schedule(self, nb_file, run_at=None, last_run=None):
        """
        Schedule a processing of a notebook as per the interval
        specified on the job script
        """
        meta = self.metadata(nb_file)
        attrs = meta.attributes
        # get/set run-at spec
        config = attrs.get('config', {})
        # see what we have as a schedule
        # -- a dictionary of JobSchedule
        if isinstance(run_at, dict):
            run_at = self.Schedule(**run_at).cron
        if isinstance(run_at, str):
            run_at = self.Schedule(run_at).cron
        # -- a JobSchedule
        if isinstance(run_at, self.Schedule):
            run_at = run_at.cron
        # -- nothing, may we have it on the job's config already
        if not run_at:
            interval = config.get('run-at')
        else:
            # store the new schedule
            config['run-at'] = run_at
            interval = run_at
        if not interval:
            # if we don't have a run-spec, return without scheduling
            raise ValueError('no run-at specification provided, cannot schedule')
        # get last time the job was run
        if last_run is None:
            job_runs = attrs.get('job_runs')
            if job_runs:
                last_run = job_runs[-1]['ts']
            else:
                last_run = datetime.datetime.now()
        # calculate next run time
        iter_next = croniter(interval, last_run)
        run_at = iter_next.get_next(datetime.datetime)
        # store next scheduled run
        triggers = attrs['triggers'] = attrs.get('triggers', [])
        # set up a schedule event
        scheduled_run = {
            'event-kind': 'scheduled',
            'event': run_at.isoformat(),
            'run-at': run_at,
            'status': 'PENDING'
        }
        # search for existing trigger, only add if not existing yet
        for cur in triggers:
            if cur.get('status') != 'PENDING':
                continue
            if scheduled_run['run-at'] == cur.get('run-at'):
                break
        else:
            triggers.append(scheduled_run)
        attrs['config'] = config
        return meta.save()

    def get_schedule(self, name, only_pending=False):
        """
        return the cron schedule and corresponding triggers

        Args:
            name (str): the name of the job

        Returns:
            tuple of (run_at, triggers)

            run_at (str): the cron spec, None if not scheduled
            triggers (list): the list of triggers
        """
        meta = self.metadata(name)
        attrs = meta.attributes
        config = attrs.get('config')
        triggers = attrs.get('triggers', [])
        if only_pending:
            triggers = [trigger for trigger in triggers
                        if  trigger['status'] == 'PENDING']
        if config and 'run-at' in config:
            run_at = config.get('run-at')
        else:
            run_at = None
        return run_at, triggers

    def drop_schedule(self, name):
        """
        Drop an existing schedule, if any

        This will drop any existing schedule and any pending triggers of
        event-kind 'scheduled'.

        Args:
            name (str): the name of the job

        Returns:
            Metadata
        """
        meta = self.metadata(name)
        attrs = meta.attributes
        config = attrs.get('config')
        triggers = attrs.get('triggers')
        if 'run-at' in config:
            del config['run-at']
        for trigger in triggers:
            if trigger['event-kind'] == 'scheduled' and trigger['status'] == 'PENDING':
                trigger['status'] = 'CANCELLED'
        return meta.save()





from inspect import isfunction
from unittest import TestCase, skip

from omegaml import Omega
from omegaml.backends.tensorflow.tfestimatormodel import TFEstimatorModelBackend, TFEstimatorModel
from omegaml.backends.virtualobj import virtualobj
from omegaml.tests.util import OmegaTestMixin, tf_perhaps_eager_execution

import tensorflow as tf


def make_data():
    # read data from csv
    import numpy as np
    import pandas as pd
    columns = ['f1', 'f2', 'f3', 'f4', 'f5']
    train_data = pd.DataFrame(np.random.random_sample((20, 5)),
                              columns=columns).astype(int)
    test_data = pd.DataFrame(np.random.random_sample((20, 5)),
                             columns=columns).astype(int)
    # separate train data
    train_x = train_data[columns[:-1]]
    train_y = train_data.loc[:, columns[-1]]

    # separate test data
    test_x = test_data[columns[:-1]]
    test_y = test_data.loc[:, columns[-1]]
    return train_x, train_y, test_x, test_y


def make_estimator_fn():
    # this is to ensure we get a serializable function
    def make_estimator(model_dir=None):
        import tensorflow as tf
        feature_columns = [tf.feature_column.numeric_column(key=key)
                           for key in ['f1', 'f2', 'f3', 'f4']]
        classifier = tf.estimator.LinearClassifier(feature_columns=feature_columns,
                                                   n_classes=3, model_dir=model_dir)
        return classifier

    return make_estimator


def make_input_fn():
    # create classifier and save untrained
    # we need to use a custom input_fn as the default won't be able to figure
    # out column names from numpy inputs
    def input_fn(mode, X, Y=None, batch_size=1):
        import tensorflow as tf
        X = {
            'f{}'.format(i + 1): X[:, i] for i in range(X.shape[1])
        }
        return tf.estimator.inputs.numpy_input_fn(x=X, y=Y, num_epochs=1, shuffle=False)

    return input_fn


class TFEstimatorModelBackendTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.om = Omega()
        self.om.models.register_backend(TFEstimatorModelBackend.KIND, TFEstimatorModelBackend)
        self.clean()
        tf_perhaps_eager_execution()

    def test_fit_predict(self):
        import tensorflow as tf
        om = self.om
        # create classifier
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        train_x, train_y, test_x, test_y = make_data()
        classifier = estmdl.fit(train_x, train_y)
        self.assertIsInstance(classifier, tf.estimator.LinearClassifier)
        # score
        score = estmdl.score(test_x, test_y)
        self.assertIsInstance(score, dict)
        self.assertIn('accuracy', score)
        self.assertIn('loss', score)
        # predict
        predict = [v for v in estmdl.predict(test_x)]
        self.assertIsInstance(predict, list)
        self.assertIn('logits', predict[0])
        self.assertIn('probabilities', predict[0])
        self.assertIn('classes', predict[0])

    def test_fit_predict_from_numpy(self):
        import tensorflow as tf
        om = self.om
        # note we use a custom input_fn
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn(), input_fn=make_input_fn())
        train_x, train_y, test_x, test_y = make_data()
        # create a feature dict from a numpy array
        train_x = train_x.values  # numpy
        train_y = train_y.values
        test_x = test_x.values
        classifier = estmdl.fit(train_x, train_y)
        self.assertIsInstance(classifier, tf.estimator.LinearClassifier)
        # score
        score = estmdl.score(test_x, test_y)
        self.assertIsInstance(score, dict)
        self.assertIn('accuracy', score)
        self.assertIn('loss', score)
        # predict
        predict = [v for v in estmdl.predict(test_x)]
        self.assertIsInstance(predict, list)
        self.assertIn('logits', predict[0])
        self.assertIn('probabilities', predict[0])
        self.assertIn('classes', predict[0])

    def test_save_load_unfitted(self):
        om = self.om
        # create classifier and save
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        meta = om.models.put(estmdl, 'estimator-model')
        # restore and use
        estmdl_r = om.models.get('estimator-model')
        self.assertIsInstance(estmdl_r, estmdl.__class__)
        train_x, train_y, test_x, test_y = make_data()
        estmdl_r.fit(train_x, train_y)
        predict = [v for v in estmdl.predict(test_x)]
        self.assertIsInstance(predict, list)
        self.assertIn('logits', predict[0])
        self.assertIn('probabilities', predict[0])
        self.assertIn('classes', predict[0])

    def test_save_load_estimator_model(self):
        import tensorflow as tf
        om = self.om
        # create classifier and save
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        meta = om.models.put(estmdl, 'estimator-model')
        # restore and use
        estmdl_r = om.models.get('estimator-model')
        # check we have a restored instance
        self.assertIsNot(estmdl_r, estmdl)
        self.assertNotEqual(estmdl.estimator.model_dir, estmdl_r.estimator.model_dir)
        self.assertIsInstance(estmdl_r, estmdl.__class__)
        self.assertIsNot(estmdl_r.estimator_fn, make_estimator_fn())
        self.assertTrue(isfunction(estmdl.estimator_fn))
        self.assertIsInstance(estmdl_r.estimator_fn(), tf.estimator.Estimator)

    def test_save_load_fitted(self):
        import numpy as np
        om = self.om
        # create classifier and save
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        train_x, train_y, test_x, test_y = make_data()
        estmdl.fit(train_x, train_y)
        predict = [v for v in estmdl.predict(test_x)]
        meta = om.models.put(estmdl, 'estimator-model')
        # restore and use
        estmdl_r = om.models.get('estimator-model')
        self.assertIsInstance(estmdl_r, estmdl.__class__)
        predict_r = [v for v in estmdl_r.predict(test_x)]
        self.assertIsInstance(predict_r, list)
        # make sure the model was actually reloaded from a new model_dir
        self.assertNotEqual(estmdl.estimator.model_dir, estmdl_r.estimator.model_dir)
        self.assertTrue(np.allclose(predict_r[0]['logits'], predict[0]['logits']))
        self.assertTrue(np.allclose(predict_r[0]['probabilities'], predict[0]['probabilities']))

    def test_save_load_fitted_estimator(self):
        import numpy as np
        om = self.om
        # create classifier and save
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        train_x, train_y, test_x, test_y = make_data()
        estmdl.fit(train_x, train_y)
        predict = [v for v in estmdl.predict(test_x)]
        meta = om.models.put(TFEstimatorModel(estimator_fn=make_estimator_fn(),
                                              model=estmdl.estimator), 'estimator-model')
        # restore and use
        estmdl_r = om.models.get('estimator-model')
        self.assertIsInstance(estmdl_r, estmdl.__class__)
        predict_r = [v for v in estmdl_r.predict(test_x)]
        self.assertIsInstance(predict_r, list)
        # make sure the model was actually reloaded from a new model_dir
        self.assertNotEqual(estmdl.estimator.model_dir, estmdl_r.estimator.model_dir)
        self.assertTrue(np.allclose(predict_r[0]['logits'], predict[0]['logits']))
        self.assertTrue(np.allclose(predict_r[0]['probabilities'], predict[0]['probabilities']))

    def test_save_load_fitted_inerror(self):
        import numpy as np
        om = self.om
        # create classifier and save untrained
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        om.models.put(estmdl, 'estimator-model')
        # create dataasets
        train_x, train_y, test_x, test_y = make_data()
        estmdl.fit(train_x, train_y)
        predict = [v for v in estmdl.predict(test_x)]
        # this effectively resets the model to untrained state
        # i.e. it can be reinstantiated but will not be trained
        estmdl._estimator = None
        meta = om.models.put(estmdl, 'estimator-model')
        # restore and use
        estmdl_r = om.models.get('estimator-model')
        self.assertIsInstance(estmdl_r, estmdl.__class__)
        predict_r = [v for v in estmdl_r.predict(test_x)]
        self.assertIsInstance(predict_r, list)
        # make sure the model was actually reloaded from a new model_dir
        self.assertNotEqual(estmdl.estimator.model_dir, estmdl_r.estimator.model_dir)
        # since we get back an unfitted model, predictions should be way off
        self.assertFalse(np.allclose(predict_r[0]['logits'], predict[0]['logits']))
        self.assertFalse(np.allclose(predict_r[0]['probabilities'], predict[0]['probabilities']))

    def test_runtime_fit(self):
        import pandas as pd
        om = self.om
        # create classifier and save untrained, note we use the default input_fn
        # provided by TFEstimatorModel as it deals easily with DataFrames
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        train_x, train_y, test_x, test_y = make_data()
        om.datasets.put(train_x, 'train_x', append=False)
        om.datasets.put(train_y, 'train_y', append=False)
        om.models.put(estmdl, 'estimator-model')
        meta = om.runtime.model('estimator-model').fit('train_x', 'train_y').get()
        self.assertTrue(meta.startswith('<Metadata:'))
        # predict using fitted model in runtime
        om.datasets.put(test_x, 'test_x', append=False)
        om.datasets.put(test_y, 'test_y', append=False)
        result = om.runtime.model('estimator-model').predict('test_x').get()
        self.assertIsInstance(result, pd.DataFrame)

    def test_runtime_predict_from_numpy(self):
        import pandas as pd
        om = self.om
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn(), input_fn=make_input_fn())
        train_x, train_y, test_x, test_y = make_data()
        train_x = train_x.values  # numpy
        train_y = train_y.values
        test_x = test_x.values
        om.datasets.put(train_x, 'train_x')
        om.datasets.put(train_y, 'train_y')
        om.datasets.put(test_x, 'test_x')
        om.models.put(estmdl, 'estimator-model')
        meta = om.runtime.model('estimator-model').fit('train_x', 'train_y').get()
        self.assertTrue(meta.startswith('<Metadata:'))
        # predict using fitted model in runtime
        om.datasets.put(test_x, 'test_x', append=False)
        om.datasets.put(test_y, 'test_y', append=False)
        result = om.runtime.model('estimator-model').predict('test_x').get()
        self.assertIsInstance(result, pd.DataFrame)

    @skip("not supported yet as we don't have a good way to store feature dicts")
    def test_runtime_predict_from_numpy_default_inputfn(self):
        import pandas as pd
        om = self.om
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        train_x, train_y, test_x, test_y = make_data()

        def as_features(x, names=None):
            # convert dataframe to feature vectors suitable for numpy_inputfn
            cols = range(x.shape[1])
            x = x.values
            features = {
                'col'.format(i + 1): x[:, i] for i, col in zip(cols, names)
            }
            return features

        train_x = as_features(train_x, train_x.columns)
        train_y = train_y.values
        test_x = as_features(test_x, test_x.columns)
        om.datasets.put(train_x, 'train_x', append=False)
        om.datasets.put(train_y, 'train_y', append=False)
        om.datasets.put(test_x, 'test_x', append=False)
        om.models.put(estmdl, 'estimator-model')
        meta = om.runtime.model('estimator-model').fit('train_x', 'train_y').get()
        self.assertTrue(meta.startswith('<Metadata:'))
        # predict using fitted model in runtime
        om.datasets.put(test_x, 'test_x', append=False)
        om.datasets.put(test_y, 'test_y', append=False)
        result = om.runtime.model('estimator-model').predict('test_x').get()
        self.assertIsInstance(result, pd.DataFrame)

    def test_runtime_score(self):
        import pandas as pd
        om = self.om
        # create classifier and save untrained
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        train_x, train_y, test_x, test_y = make_data()
        om.datasets.put(train_x, 'train_x', append=False)
        om.datasets.put(train_y, 'train_y', append=False)
        om.models.put(estmdl, 'estimator-model')
        meta = om.runtime.model('estimator-model').fit('train_x', 'train_y').get()
        self.assertTrue(meta.startswith('<Metadata:'))
        # predict using fitted model in runtime
        om.datasets.put(test_x, 'test_x', append=False)
        om.datasets.put(test_y, 'test_y', append=False)
        result = om.runtime.model('estimator-model').score('test_x', 'test_y').get()
        self.assertIsInstance(result, pd.Series)
        self.assertAlmostEqual(result['accuracy'], 1.0)

    def test_predict_from_objecthandler(self):
        import tensorflow as tf
        om = self.om

        @virtualobj
        def train_xy_fn(Xname=None, Yname=None, **kwargs):
            import tensorflow as tf
            import omegaml as om
            X = om.datasets.get(Xname)
            Y = om.datasets.get(Yname)
            dataset = tf.estimator.inputs.pandas_input_fn(X, Y, shuffle=True)
            return dataset

        @virtualobj
        def test_x_fn(Xname=None, **kwargs):
            import tensorflow as tf
            import omegaml as om
            X = om.datasets.get(Xname)
            dataset = tf.estimator.inputs.pandas_input_fn(X, shuffle=False)
            return dataset

        # create classifier
        estmdl = TFEstimatorModel(estimator_fn=make_estimator_fn())
        train_x, train_y, test_x, test_y = make_data()
        # store actual data
        om.datasets.put(train_x, 'train_x')
        om.datasets.put(train_y, 'train_y')
        om.datasets.put(test_x, 'test_x')
        # make these virtual objects to return (the input fn that returns a) tf.data.Dataset
        om.datasets.put(train_xy_fn, 'train_data')
        om.datasets.put(test_x_fn, 'test_dataset')
        # use these datasets as an input fn
        # -- why can't we pass the Dataset directly? All objects must be executed on the same
        #    graph wich is only created by estimator.fit() before calling the input_fn.
        dataset = om.datasets.get('train_data', Xname='train_x', Yname='train_y')
        classifier = estmdl.fit(input_fn=dataset)
        self.assertIsInstance(classifier, tf.estimator.LinearClassifier)
        # predict directly
        dataset = om.datasets.get('test_dataset', Xname='test_x')
        predict = [v for v in estmdl.predict(input_fn=dataset)]
        self.assertIsInstance(predict, list)
        self.assertIn('logits', predict[0])
        self.assertIn('probabilities', predict[0])
        self.assertIn('classes', predict[0])
        # do the same thing on the runtime
        om.models.put(estmdl, 'estimator-model')
        meta = om.runtime.model('estimator-model').fit('train_data{Xname=train_x,Yname=train_y}').get()
        self.assertTrue(meta.startswith('<Metadata:'))



import re

import six

from omegaml.backends.virtualobj import VirtualObjectBackend


class VirtualObjectMixin(object):
    """
    process virtual objects

    This checks if an object is a VirtualObject and if so
    retrieves the handler and processes it.
    """

    def __init__(self):
        self.__meta = None

    def _isvirtual(self, name):
        self.__meta = self.metadata(name)
        return (self.__meta is not None and
                self.__meta.kind == VirtualObjectBackend.KIND)

    def _getvirtualobjfn(self, name):
        virtualobjfn = super(VirtualObjectMixin, self).get(name)
        return virtualobjfn

    def _resolve_realname(self, name, kwargs):
        # parse names in the format name{a=b,c=d} to name and update kwargs
        if name and all(c in name for c in '{}'):
            rx = r"(\w*=\w*)*"
            real_name, fn_kwargs = name.split('{', 1)
            fn_kwargs = fn_kwargs.split('}')[0]
            matches = (v.split('=') for v in re.findall(rx, fn_kwargs) if v)
            fn_kwargs = {
                k: v for k, v in matches
            }
            kwargs.update(fn_kwargs)
            name = real_name
        return name, kwargs

    def get(self, name, **kwargs):
        raw = kwargs.get('raw', False)
        name, kwargs = self._resolve_realname(name, kwargs)
        if not raw and self._isvirtual(name):
            handler = self._getvirtualobjfn(name)
            result = handler(method='get', meta=self.__meta, store=self, **kwargs)
        else:
            result = super(VirtualObjectMixin, self).get(name, **kwargs)
        return result

    def put(self, obj, name, replace=False, attributes=None, **kwargs):
        name, kwargs = self._resolve_realname(name, kwargs)
        if not replace and self._isvirtual(name):
            result = self._getvirtualobjfn(name)(data=obj, method='put',
                                                 meta=self.__meta, store=self, **kwargs)
        else:
            result = super(VirtualObjectMixin, self).put(obj, name, attributes=attributes, **kwargs)
        return result

    def drop(self, name, force=False, version=-1):
        if self._isvirtual(name):
            try:
                handler = self._getvirtualobjfn(name)
                result = handler(method='drop', meta=self.__meta, store=self, force=force)
            except:
                if not force:
                    raise
            else:
                if result is False:
                    return False
        return super(VirtualObjectMixin, self).drop(name, force=force)




import six
from celery import Task
from kombu.serialization import registry
from kombu.utils import cached_property


class EagerSerializationTaskMixin(object):
    # ensure eager tasks are being serialized to capture serialization errors
    # adopted from https://github.com/celery/celery/issues/4008#issuecomment-330292405
    abstract = True

    @cached_property
    def _not_eager(self):
        app = self._app
        return not app.conf.CELERY_ALWAYS_EAGER

    @cached_property
    def _task_serializer(self):
        app = self._app
        return app.conf.CELERY_TASK_SERIALIZER

    def apply_async(self, args=None, kwargs=None, *args_, **kwargs_):
        if self._not_eager:
            return super(EagerSerializationTaskMixin, self).apply_async(args=args, kwargs=kwargs, *args_, **kwargs_)
        # only execute if eager
        args, kwargs = self._eager_serialize_args(args=args, kwargs=kwargs, **kwargs_)
        result = super(EagerSerializationTaskMixin, self).apply_async(args=args, kwargs=kwargs, **kwargs_)
        result = self._eager_serialize_result(result, **kwargs_)
        return result

    def _eager_serialize_args(self, args=None, kwargs=None, **kwargs_):
        # Perform a noop serialization backtrip to assert args and kwargs
        # will be serialized appropriately when an async call through kombu
        # is actually performed. This is done to make sure we catch the
        # serializations errors with our test suite which runs with the
        # CELERY_ALWAYS_EAGER setting set to True. See the following Celery
        # issue for details https://github.com/celery/celery/issues/4008.
        app = self._app
        producer = kwargs_.get('producer') if kwargs else None
        with app.producer_or_acquire(producer) as producer:
            serializer = kwargs_.get('serializer', producer.serializer) or self._task_serializer
            registry.enable(serializer)
            args_content_type, args_content_encoding, args_data = registry.dumps(args, serializer)
            kwargs_content_type, kwargs_content_encoding, kwargs_data = registry.dumps(kwargs, serializer)
            args = registry.loads(args_data, args_content_type, args_content_encoding)
            kwargs = registry.loads(kwargs_data, kwargs_content_type, kwargs_content_encoding)
        return args, kwargs

    def _eager_serialize_result(self, result, **kwargs_):
        app = self._app
        producer = kwargs_.get('producer') if kwargs_ else None
        result_value = result._result
        with app.producer_or_acquire(producer) as producer:
            serializer = kwargs_.get('serializer', producer.serializer) or self._task_serializer
            registry.enable(serializer)
            dtype, encoding, data = registry.dumps(result_value, serializer)
            result._result = registry.loads(data, dtype, encoding)
        return result


class OmegamlTask(EagerSerializationTaskMixin, Task):
    abstract = True

    def __init__(self, *args, **kwargs):
        super(OmegamlTask, self).__init__(*args, **kwargs)
        self._om = None

    @property
    def om(self):
        # TODO do some more intelligent caching, i.e. by client/auth
        if self._om is None:
            from omegaml import get_omega_for_task
            bucket = self.request.kwargs.pop('__bucket', None)
            self._om = get_omega_for_task(self)[bucket]
        return self._om

    def get_delegate(self, name, kind='models'):
        get_delegate_provider = getattr(self.om, kind)
        return get_delegate_provider.get_backend(name, data_store=self.om.datasets)

    @property
    def delegate_args(self):
        return self.request.args

    @property
    def delegate_kwargs(self):
        return {k: v for k, v in six.iteritems(self.request.kwargs) if not k.startswith('__')}

    def reset(self):
        # ensure next call will start over and get a new om instance
        self._om = None

    def on_failure(self, *args, **kwargs):
        self.reset()
        return super().on_failure(*args, **kwargs)

    def on_retry(self, *args, **kwargs):
        self.reset()
        return super().on_retry(*args, **kwargs)

    def on_success(self, *args, **kwargs):
        self.reset()
        return super().on_success(*args, **kwargs)


def get_dataset_representations(items):
    """
    returns dict with x and y datasets
    """
    results = {}
    results['Xname'] = items.get('Xname')
    results['Yname'] = items.get('Yname')
    return results


def sanitized(value):
    # fix because local Metadata object cannot be pickled
    if getattr(type(value), '__name__', None) == 'Metadata':
        value = repr(value)
    return value



from __future__ import absolute_import

import glob
from zipfile import ZipFile, ZIP_DEFLATED

import datetime
import os
import tempfile
from mongoengine.fields import GridFSProxy
from shutil import rmtree
from sklearn.base import BaseEstimator
from sklearn.model_selection import GridSearchCV

from omegaml.backends.basemodel import BaseModelBackend
from omegaml.documents import MDREGISTRY
from omegaml.util import reshaped, gsreshaped


class ScikitLearnBackend(BaseModelBackend):
    """
    OmegaML backend to use with ScikitLearn
    """

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, BaseEstimator)

    def _package_model(self, model, filename):
        """
        Dumps a model using joblib and packages all of joblib files into a zip
        file
        """
        import joblib
        lpath = tempfile.mkdtemp()
        fname = os.path.basename(filename)
        mklfname = os.path.join(lpath, fname)
        zipfname = os.path.join(self.model_store.tmppath, fname)
        joblib.dump(model, mklfname, protocol=4)
        with ZipFile(zipfname, 'w', compression=ZIP_DEFLATED) as zipf:
            for part in glob.glob(os.path.join(lpath, '*')):
                zipf.write(part, os.path.basename(part))
        rmtree(lpath)
        return zipfname

    def _extract_model(self, packagefname):
        """
        Loads a model using joblib from a zip file created with _package_model
        """
        import joblib
        lpath = tempfile.mkdtemp()
        fname = os.path.basename(packagefname)
        mklfname = os.path.join(lpath, fname)
        with ZipFile(packagefname) as zipf:
            zipf.extractall(lpath)
        model = joblib.load(mklfname)
        rmtree(lpath)
        return model

    def get_model(self, name, version=-1):
        """
        Retrieves a pre-stored model
        """
        filename = self.model_store._get_obj_store_key(name, '.omm')
        packagefname = os.path.join(self.model_store.tmppath, name)
        dirname = os.path.dirname(packagefname)
        try:
            os.makedirs(dirname)
        except OSError:
            # OSError is raised if path exists already
            pass
        outf = self.model_store.fs.get_version(filename, version=version)
        with open(packagefname, 'wb') as zipf:
            zipf.write(outf.read())
        model = self._extract_model(packagefname)
        return model

    def put_model(self, obj, name, attributes=None):
        """
        Packages a model using joblib and stores in GridFS
        """
        zipfname = self._package_model(obj, name)
        with open(zipfname, 'rb') as fzip:
            fileid = self.model_store.fs.put(
                fzip, filename=self.model_store._get_obj_store_key(name, 'omm'))
            gridfile = GridFSProxy(grid_id=fileid,
                                   db_alias='omega',
                                   collection_name=self.model_store.bucket)
        return self.model_store._make_metadata(
            name=name,
            prefix=self.model_store.prefix,
            bucket=self.model_store.bucket,
            kind=MDREGISTRY.SKLEARN_JOBLIB,
            attributes=attributes,
            gridfile=gridfile).save()

    def predict(
          self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        data = self.data_store.get(Xname)
        model = self.model_store.get(modelname)
        result = model.predict(reshaped(data), **kwargs)
        if pure_python:
            result = result.tolist()
        if rName:
            meta = self.data_store.put(result, rName)
            result = meta
        return result

    def predict_proba(
          self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        data = self.data_store.get(Xname)
        model = self.model_store.get(modelname)
        result = model.predict_proba(reshaped(data), **kwargs)
        if pure_python:
            result = result.tolist()
        if rName:
            meta = self.data_store.put(result, rName)
            result = meta
        return result

    def fit(self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
        model = self.model_store.get(modelname)
        X, metaX = self.data_store.get(Xname), self.data_store.metadata(Xname)
        Y, metaY = None, None
        if Yname:
            Y, metaY = (self.data_store.get(Yname),
                        self.data_store.metadata(Yname))
        model.fit(reshaped(X), reshaped(Y), **kwargs)
        # store information required for retraining
        model_attrs = {
            'metaX': metaX.to_mongo(),
            'metaY': metaY.to_mongo() if metaY is not None else None,
        }
        try:
            import sklearn
            model_attrs['scikit-learn'] = sklearn.__version__
        except:
            model_attrs['scikit-learn'] = 'unknown'
        meta = self.model_store.put(model, modelname, attributes=model_attrs)
        return meta

    def partial_fit(
          self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
        model = self.model_store.get(modelname)
        X, metaX = self.data_store.get(Xname), self.data_store.metadata(Xname)
        Y, metaY = None, None
        if Yname:
            Y, metaY = (self.data_store.get(Yname),
                        self.data_store.metadata(Yname))
        model.partial_fit(reshaped(X), reshaped(Y), **kwargs)
        # store information required for retraining
        model_attrs = {
            'metaX': metaX.to_mongo(),
            'metaY': metaY.to_mongo() if metaY is not None else None,
        }
        try:
            import sklearn
            model_attrs['scikit-learn'] = sklearn.__version__
        except:
            model_attrs['scikit-learn'] = 'unknown'
        meta = self.model_store.put(model, modelname, attributes=model_attrs)
        return meta

    def score(
          self, modelname, Xname, Yname, rName=None, pure_python=True,
          **kwargs):
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname)
        result = model.score(reshaped(X), reshaped(Y), **kwargs)
        if rName:
            meta = self.model_store.put(result, rName)
            result = meta
        return result

    def fit_transform(
          self, modelname, Xname, Yname=None, rName=None, pure_python=True,
          **kwargs):
        model = self.model_store.get(modelname)
        X, metaX = self.data_store.get(Xname), self.data_store.metadata(Xname)
        Y, metaY = None, None
        if Yname:
            Y, metaY = (self.data_store.get(Yname),
                        self.data_store.metadata(Yname))
        result = model.fit_transform(reshaped(X), reshaped(Y), **kwargs)
        # store information required for retraining
        model_attrs = {
            'metaX': metaX.to_mongo(),
            'metaY': metaY.to_mongo() if metaY is not None else None
        }
        try:
            import sklearn
            model_attrs['scikit-learn'] = sklearn.__version__
        except:
            model_attrs['scikit-learn'] = 'unknown'
        meta = self.model_store.put(model, modelname, attributes=model_attrs)
        if pure_python:
            result = result.tolist()
        if rName:
            meta = self.data_store.put(result, rName)
        result = meta
        return result

    def transform(self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)
        result = model.transform(reshaped(X), **kwargs)
        if pure_python:
            result = result.tolist()
        if rName:
            meta = self.data_store.put(result, rName)
            result = meta
        return result

    def decision_function(self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        model = self.model_store.get(modelname)
        X = self.data_store.get(Xname)
        result = model.decision_function(reshaped(X), **kwargs)
        if pure_python:
            result = result.tolist()
        if rName:
            meta = self.data_store.put(result, rName)
            result = meta
        return result

    def gridsearch(self, modelname, Xname, Yname, rName=None,
                   parameters=None, pure_python=True, **kwargs):
        model, meta = self.model_store.get(modelname), self.model_store.metadata(modelname)
        X = self.data_store.get(Xname)
        if Yname:
            y = self.data_store.get(Yname)
        else:
            y = None
        gs_model = GridSearchCV(cv=5, estimator=model, param_grid=parameters, **kwargs)
        gs_model.fit(X, gsreshaped(y))
        nowdt = datetime.datetime.now()
        if rName:
            gs_modelname = rName
        else:
            gs_modelname = '{}.{}.gs'.format(modelname, nowdt.isoformat())
        gs_meta = self.model_store.put(gs_model, gs_modelname)
        attributes = meta.attributes
        if not 'gridsearch' in attributes:
            attributes['gridsearch'] = []
        attributes['gridsearch'].append({
            'datetime': nowdt,
            'Xname': Xname,
            'Yname': Yname,
            'gsModel': gs_modelname,
        })
        meta.save()
        return meta



from time import sleep
from unittest import TestCase

import os

import nbformat

from omegaml import Omega
from omegaml.tests.cli.scenarios import CliTestScenarios
from omegaml.tests.util import OmegaTestMixin
from omegaml.util import temp_filename


class CliJobsTest(CliTestScenarios, OmegaTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.om = Omega()
        self.clean()

    def test_cli_jobs_list(self):
        self.cli('jobs list')
        self.assertLogContains('info', '[]')
        self.create_job('test')
        self.cli('jobs list')
        self.assertLogContains('info', 'test.ipynb')

    def test_cli_jobs_get(self):
        self.create_job('test')
        fn = temp_filename(ext='ipynb')
        self.cli(f'jobs get test {fn}')
        self.assertTrue(os.path.exists(fn))
        with open(fn, 'r') as fin:
            nb = nbformat.read(fin, as_version=4)
        self.assertIn('cells', nb)
        self.assertIn('hello', nb['cells'][0]['source'])

    def test_cli_jobs_put(self):
        self.assertNotIn('test', self.om.jobs.list())
        cells = []
        code = "print 'hello'"
        cells.append(nbformat.v4.new_code_cell(source=code))
        notebook = nbformat.v4.new_notebook(cells=cells)
        fn = temp_filename(ext='ipynb')
        with open(fn, 'w') as fout:
            nbformat.write(notebook, fout, version=4)
        self.cli(f'jobs put {fn} test')
        self.assertLogContains('info', 'Metadata')
        self.assertLogContains('info', 'name=test')
        self.assertIn('test.ipynb', self.om.jobs.list())

    def test_cli_jobs_metadata(self):
        self.create_job('test')
        self.cli('jobs metadata test')
        expected = self.pretend_log('"kind": "script.ipynb"')
        self.assertLogContains('info', expected)
        expected = self.pretend_log('"name": "test.ipynb"')
        self.assertLogContains('info', expected)

    def test_cli_jobs_schedule(self):
        self.create_job('test')
        # test there is an input prompt
        self.cli('jobs schedule test now', askfn=lambda *a, **kw: 'y')
        self.assertLogContains('info', 'test will be scheduled to run Every minute')
        # test there is no input prompt
        self.cli('jobs schedule test now -q', new_start=True)
        self.assertLogContains('info', 'Currently test is scheduled at Every minute')
        self.assertLogContains('info', 'test is scheduled to run next at')
        # test scheduled can be shown
        self.cli('jobs schedule test show', new_start=True)
        self.assertLogContains('info', 'test is scheduled to run next at')
        # test we can delete the job schedule
        # test there is an input prompt
        self.cli('jobs schedule test delete', askfn=lambda *a, **kw: 'y', new_start=True)
        self.cli('jobs schedule test show', new_start=True)
        self.assertLogContains('info', 'Currently test is not scheduled')
        # test we can see executed jobs
        self.cli('jobs schedule test now', askfn=lambda *a, **kw: 'y')
        meta = self.om.jobs.metadata('test')
        triggers = meta.attributes['triggers']
        trigger = triggers[-1]
        kwargs = dict(now=trigger['run-at'])
        self.om.runtime.task('omegaml.notebook.tasks.execute_scripts').apply_async(kwargs=kwargs).get()
        self.cli('jobs status test', new_start=True)
        entries =  self.get_log('info')
        # entries example - a list of args tuples passed to logger.info(*args)
        #   [('Runs:',), ('  2019-11-15 18:32:28.625000 OK ',),
        #   ('Next scheduled runs:',), ('   PENDING scheduled 2019-11-15T18:34:00',)]
        self.assertIn('Runs:', entries[0][0])
        self.assertIn('OK', entries[1][0])
        self.assertIn('Next scheduled runs:', entries[2][0])
        self.assertIn('PENDING', entries[3][0])



import datetime
from croniter import croniter
import six
from celery.schedules import crontab


class JobSchedule(object):
    """
    Produce a cron tab spec from text, time periods, or a crontab spec

    Given any specification format, can translate to a human readable
    text.

    If using time periods (minute, hour, weekday, monthday, month),
    any argument not specified defaults to '*' for this argument.

    If using text, sepearate each time part (weekday, hour, month)
    by a comma. To specify multiple times for a part, use / instead
    of comma.

    Examples:
        # using text
        JobSchedule('friday, at 06:00/08:00/10:00')
        JobSchedule('Mondays and Fridays, at 06:00')
        JobSchedule('every 5 minutes, on weekends, in april')

        # using time periods
        JobSchedule(weekday='mon-fri', at='06:05,12:05')
        JobSchedule(weekday='mon-fri', at='06:00')
        JobSchedule(month='every 2', at='08:00', weekday='every 3')

        # using a crontab spec
        JobSchedule('05 06,12 * * mon-fri')

        # given a valid specification get human readable text or crontab format
        JobSchedule('05 06,12 * * mon-fri').text
        JobSchedule('Mondays and Fridays, at 06:00').cron

    Args:
        text: the natural language specification, with time parts
              separated by comma
        at: the hh:mm specification, equal to hour=hh, minute=mm
        minute: run on 0-59th minute in every specified hour
        hour: run on 0-23th hour on every specified day
        weekday: run on 0-6th day in every week (0 is Sunday),
                 can also be specified as mon/tue/wed/thu/fri/sat/sun
        monthday: run on 1-31th day in every month
        month: run on 1-12th day of every year

    Raises:
        ValueError if the given specification is not correct

    Returns:
        the cron specificiation, use for run_at in self.schedule()
    """

    def __init__(self, text=None, minute='*', hour='*', weekday='*',
                 monthday='*', month='*', at=None):
        # if we get text, attempt to convert
        if text:
            self.sched = self._convert_text(text).sched
            return
        # no text, but time periods
        # get times
        if at:
            # 06:00 => hour=6 / minute=00
            # 06:00,12:00 => hour=6,12 / minute=00
            hours, minutes = [], []
            for att in at.split(','):
                h, m = att.split(':')
                hours.append(h)
                minutes.append(m)
            hour = ','.join(sorted(set(hours)))
            minute = ','.join(sorted(set(minutes)))
        # every n => */n
        minute = self._expand_every(minute)
        hour = self._expand_every(hour)
        weekday = self._expand_every(weekday)
        monthday = self._expand_every(monthday)
        month = self._expand_every(month)
        # long to short, e.g. friday => fri
        weekday = self._convert_weekdays(weekday)
        # month to number, e.g. january = 1
        month = self._convert_months(month)
        # get a cron spec
        self.sched = crontab(minute=minute,
                             hour=hour,
                             day_of_month=monthday,
                             day_of_week=weekday,
                             month_of_year=month)
        # make sure the spec can be processed by croniter
        if not croniter.is_valid(self.cron):
            raise ValueError("{cronspec} is not a valid schedule")

    @classmethod
    def from_cron(cls, cronspec):
        (minute, hour, monthday, month, weekday) = cronspec.split(' ')
        return JobSchedule(minute=minute, hour=hour, monthday=monthday,
                           month=month, weekday=weekday)

    @classmethod
    def from_text(cls, text):
        return JobSchedule(text=text)

    @property
    def cron(self):
        # adopted from https://docs.celeryproject.org/en/latest/_modules/celery/schedules.html#schedule
        cron_repr = ('{0._orig_minute} {0._orig_hour} {0._orig_day_of_month} '
                     '{0._orig_month_of_year} {0._orig_day_of_week}')
        return cron_repr.format(self.sched)

    @property
    def text(self):
        from cron_descriptor import get_description
        return get_description(self.cron)

    def next_times(self, n=None, last_run=None):
        iter_next = croniter(self.cron, start_time=last_run)
        while n > 0 or n is None:
            yield iter_next.get_next(datetime.datetime)
            n -= 1 if n > 0 else None

    def __repr__(self):
        return 'JobSchedule(cron={}, text={})'.format(self.cron, self.text)

    def _convert_weekdays(self, v):
        # convert full name weekdays to short
        days = dict([('monday', 'mon'),
                     ('tuesday', 'tue'),
                     ('wednesday', 'wed'),
                     ('thursday', 'thu'),
                     ('friday', 'fri'),
                     ('saturday', 'sat'),
                     ('sunday', 'sun'),
                     ('weekday', 'mon-fri'),
                     ('workday', 'mon-fri'),
                     ('working day', 'mon-fri'),
                     ('weekend', 'sat-sun'),
                     ('week-end', 'sat-sun'),
                     ('week end', 'sat-sun'),
                     ])
        v = v.lower()
        for full, short in days.items():
            v = v.replace(full, short)
        return v

    def _has_month(self, v):
        months = ('january,february,march,april,may,june,july,august,'
                  'september,october,november,december').split(',')
        long = any(m in v for m in months)
        short = any(m[0:3] in v for m in months)
        return long or short

    def _has_day(self, v):
        days = ('monday,tuesday,wednesday,thursday,friday,saturday,sunday').split(',')
        long = any(d in v for d in days)
        short = any(d[0:3] in v for d in days)
        return long or short

    def _convert_months(self, v):
        months = ('january,february,march,april,may,june,july,august,'
                  'september,october,november,december').split(',')
        if not self._has_month(v):
            return v
        v = v.lower()
        for m in months:
            # full
            start = v.find(m)
            if start > -1:
                v = v.replace(m, str(months.index(m) + 1))
            # short
            sm = m[0:3]
            start = v.find(sm)
            if start > -1:
                v = v.replace(sm, str(months.index(m) + 1))
        return v

    def _expand_every(self, v):
        # convert 'every' specs to cron-like
        if not isinstance(v, six.string_types):
            return v
        # every n(nd,rd,th) => */n
        if '1st' in v or 'first' in v:
            # 'every 1st' => 1
            v = '1'
        if v.startswith('every '):
            v = v.replace('every ', '*/')
            for order in ('nd', 'rd', 'th'):
                v = v.replace(order, '')
        else:
            # 'every' => *
            v = v.replace('every', '*')
        return v.strip()

    def _convert_text(self, text):
        # experimental natural language text to crontab
        # every friday
        specs = {}
        # ensure single whitespace
        orig_text = text
        text = ' '.join(text.split()).lower()
        # try placing commas
        text = text.replace(' ', ',')
        text = text.replace('every,', 'every ')
        text = text.replace(',and,', ' and ')
        text = text.replace(',only,', ' only ')
        text = text.replace('at,', 'at ')
        text = text.replace('on,', 'on ')
        text = text.replace('in,', 'in ')
        text = text.replace(',through,', ' through ')
        text = text.replace(',-,', ' - ')
        text = text.replace(',until,', ' until ')
        text = text.replace(',till,', ' till ')
        text = text.replace(',to,', ' to ')
        text = text.replace(',of,', ' of ')
        text = text.replace(',from,', ' from ')
        text = text.replace(',hours,', ' hours,')
        text = text.replace(',hour,', ' hour,')
        text = text.replace('day hour,', 'day,hour ')
        text = text.replace('hours,', 'hours ')
        text = text.replace('hour,', 'hour ')
        text = text.replace('working,day', 'working day')
        text = text.replace('week,day', 'week day')
        text = text.replace(',minutes', ' minutes')
        text = text.replace(',minute', ' minute')
        text = text.replace(',day', ' day')
        text = text.replace(',end', ' end')
        text = text.replace(',month,', ' month,')
        # get parts separated by comma
        parts = [part.strip() for part in text.split(',') if part]
        try:
            specs = self._parse_parts(parts)
            sched = JobSchedule(**specs)
        except:
            raise ValueError(f'Cannot parse {orig_text}, read as {parts}')
        return sched

    def _parse_parts(self, parts):
        specs = {}
        for part in parts:
            # Monday and Friday => Monday/Friday
            part = part.replace(' and ', ',')
            part = part.replace('days', 'day')
            part = part.replace('hours', 'hour')
            part = part.replace('minutes', 'minute')
            part = part.replace('ends', 'end')
            part = part.replace('on ', '')
            part = part.replace('in ', '')
            part = part.replace('/', ',')
            part = part.replace(' through ', '-')
            part = part.replace(' until ', '-')
            part = part.replace(' till ', '-')
            part = part.replace(' to ', '-')
            part = part.replace(' of ', '')
            part = part.replace('from ', '')
            if 'day' in part and 'month' in part:
                specs['monthday'] = part.replace('day', '').replace('month', '')
            elif 'month' in part:
                # 'every month', 'every 3rd month'
                specs['month'] = part.replace('month', '').strip()
            elif self._has_month(part):
                # 'january', 'every january'
                specs['month'] = self._convert_months(part).replace('every', '')
            elif 'daily' in part:
                specs['weekday'] = '*'
            elif (self._has_day(part) or 'day' in part or 'week' in part):
                # 'monday-friday', 'tuesday', 'every 3rd day'
                part = (self._convert_weekdays(part)
                        .replace('day', '')
                        .strip())
                # every mon-fri => mon-fri
                if 'every ' in part and '-' in part:
                    part = part.replace('every ', '').strip()
                specs['weekday'] = part
            elif ':' in part:
                # at 06:00
                specs['at'] = part.replace('at', '').strip()
            elif 'hour' in part:
                # every hour, hour 6, hour 6/7
                specs['hour'] = part.replace('hour', '').strip()
            elif 'minute' in part:
                # every minute, every 3rd minute, every 5 minutes, minute 6/7
                specs['minute'] = part.replace('minute', '').replace('at ', '').strip()
        return specs



from __future__ import absolute_import

import logging
from uuid import uuid4

import six

from omegaml.util import is_dataframe, settings, is_ndarray, extend_instance

logger = logging.getLogger(__file__)


class ModelMixin(object):
    def fit(self, Xname, Yname=None, **kwargs):
        """
        fit the model

        Calls :code:`.fit(X, Y, **kwargs)`. If instead of dataset names actual data
        is given, the data is stored using _fitX/fitY prefixes and a unique
        name.

        After fitting, a new model version is stored with its attributes
        fitX and fitY pointing to the datasets, as well as the sklearn
        version used.

        :param Xname: name of X dataset or data
        :param Yname: name of Y dataset or data
        :return: the model (self) or the string representation (python clients)
        """
        omega_fit = self.task('omegaml.tasks.omega_fit')
        Xname = self._ensure_data_is_stored(Xname, prefix='_fitX')
        if Yname is not None:
            Yname = self._ensure_data_is_stored(Yname, prefix='_fitY')
        return omega_fit.delay(self.modelname, Xname, Yname, **kwargs)

    def partial_fit(self, Xname, Yname=None, **kwargs):
        """
        update the model

        Calls :code:`.partial_fit(X, Y, **kwargs)`. If instead of dataset names actual 
        data  is given, the data is stored using _fitX/fitY prefixes and 
        a unique name. 

        After fitting, a new model version is stored with its attributes
        fitX and fitY pointing to the datasets, as well as the sklearn
        version used.   

        :param Xname: name of X dataset or data
        :param Yname: name of Y dataset or data
        :return: the model (self) or the string representation (python clients) 
        """
        omega_fit = self.task('omegaml.tasks.omega_partial_fit')
        Xname = self._ensure_data_is_stored(Xname, prefix='_fitX')
        if Yname is not None:
            Yname = self._ensure_data_is_stored(Yname, prefix='_fitY')
        return omega_fit.delay(self.modelname, Xname, Yname, **kwargs)

    def transform(self, Xname, rName=None, **kwargs):
        """
        transform X

        Calls :code:`.transform(X, **kwargs)`. If rName is given the result is
        stored as object rName

        :param Xname: name of the X dataset
        :param rName: name of the resulting dataset (optional)
        :return: the data returned by .transform, or the metadata of the rName
            dataset if rName was given
        """
        omega_transform = self.task('omegaml.tasks.omega_transform')
        Xname = self._ensure_data_is_stored(Xname)
        return omega_transform.delay(self.modelname, Xname,
                                     rName=rName, **kwargs)

    def fit_transform(self, Xname, Yname=None, rName=None, **kwargs):
        """
        fit & transform X

        Calls :code:`.fit_transform(X, Y, **kwargs)`. If rName is given the result is
        stored as object rName

        :param Xname: name of the X dataset
        :param Yname: name of the Y dataset
        :param rName: name of the resulting dataset (optional)
        :return: the data returned by .fit_transform, or the metadata of the rName
           dataset if rName was given
        """

        omega_fit_transform = self.task(
            'omegaml.tasks.omega_fit_transform')
        Xname = self._ensure_data_is_stored(Xname)
        if Yname is not None:
            Yname = self._ensure_data_is_stored(Yname)
        return omega_fit_transform.delay(self.modelname, Xname, Yname,
                                         rName=rName, transform=True, **kwargs)

    def predict(self, Xpath_or_data, rName=None, **kwargs):
        """
        predict

        Calls :code:`.predict(X)`. If rName is given the result is
        stored as object rName

        :param Xname: name of the X dataset
        :param rName: name of the resulting dataset (optional)
        :return: the data returned by .predict, or the metadata of the rName
            dataset if rName was given
        """
        omega_predict = self.task('omegaml.tasks.omega_predict')
        Xname = self._ensure_data_is_stored(Xpath_or_data)
        return omega_predict.delay(self.modelname, Xname, rName=rName, **kwargs)

    def predict_proba(self, Xpath_or_data, rName=None, **kwargs):
        """
        predict probabilities

        Calls :code:`.predict_proba(X)`. If rName is given the result is
        stored as object rName

        :param Xname: name of the X dataset
        :param rName: name of the resulting dataset (optional)
        :return: the data returned by .predict_proba, or the metadata of the rName
           dataset if rName was given
        """
        omega_predict_proba = self.task(
            'omegaml.tasks.omega_predict_proba')
        Xname = self._ensure_data_is_stored(Xpath_or_data)
        return omega_predict_proba.delay(self.modelname, Xname, rName=rName, **kwargs)

    def score(self, Xname, yName, rName=None, **kwargs):
        """
        calculate score

        Calls :code:`.score(X, y, **kwargs)`. If rName is given the result is
        stored as object rName

        :param Xname: name of the X dataset
        :param yName: name of the y dataset
        :param rName: name of the resulting dataset (optional)
        :return: the data returned by .score, or the metadata of the rName
           dataset if rName was given
        """
        omega_score = self.task('omegaml.tasks.omega_score')
        Xname = self._ensure_data_is_stored(Xname)
        yName = self._ensure_data_is_stored(yName)
        return omega_score.delay(self.modelname, Xname, yName, rName=rName, **kwargs)

    def decision_function(self, Xname, rName=None, **kwargs):
        """
        calculate score

        Calls :code:`.decision_function(X, y, **kwargs)`. If rName is given the result is
        stored as object rName

        :param Xname: name of the X dataset
        :param rName: name of the resulting dataset (optional)
        :return: the data returned by .score, or the metadata of the rName
           dataset if rName was given
        """
        omega_decision_function = self.task('omegaml.tasks.omega_decision_function')
        Xname = self._ensure_data_is_stored(Xname)
        return omega_decision_function.delay(self.modelname, Xname, rName=rName, **kwargs)

    def _ensure_data_is_stored(self, name_or_data, prefix='_temp'):
        if is_dataframe(name_or_data):
            name = '%s_%s' % (prefix, uuid4().hex)
            self.runtime.omega.datasets.put(name_or_data, name)
        elif is_ndarray(name_or_data):
            name = '%s_%s' % (prefix, uuid4().hex)
            self.runtime.omega.datasets.put(name_or_data, name)
        elif isinstance(name_or_data, (list, tuple, dict)):
            name = '%s_%s' % (prefix, uuid4().hex)
            self.runtime.omega.datasets.put(name_or_data, name)
        elif isinstance(name_or_data, six.string_types):
            name = name_or_data
        else:
            raise TypeError(
                'invalid type for Xpath_or_data', type(name_or_data))
        return name



from __future__ import absolute_import

from os.path import basename

import logging
import os
import six
import sys
import yaml

from omegaml.util import tensorflow_available, keras_available, module_available

# determine how we're run
is_cli_run = os.path.basename(sys.argv[0]) == 'om'
is_test_run = any(m in [basename(arg) for arg in sys.argv]
                  for m in ('unittest', 'test', 'nosetests', 'noserunner', '_jb_unittest_runner.py',
                            '_jb_nosetest_runner.py'))

#: configuration file, by default will be searched in current directory, user config or site config
OMEGA_CONFIG_FILE = os.environ.get('OMEGA_CONFIG_FILE') or 'config.yml'
#: the temp directory used by omegaml processes
OMEGA_TMP = '/tmp'
#: the fully qualified mongodb database URL, including the database name
OMEGA_MONGO_URL = (os.environ.get('OMEGA_MONGO_URL') or
                   os.environ.get('MONGO_URL') or
                   'mongodb://admin:foobar@localhost:27017/omega')
#: the collection name in the mongodb used by omegaml storage
OMEGA_MONGO_COLLECTION = 'omegaml'
#: determine if we should use SSL for mongodb and rabbitmq
OMEGA_USESSL = True if os.environ.get('OMEGA_USESSL') else False
#: additional kwargs for mongodb SSL connections
OMEGA_MONGO_SSL_KWARGS = {
    'ssl': OMEGA_USESSL,
}
#: if set forces eager execution of runtime tasks
OMEGA_LOCAL_RUNTIME = os.environ.get('OMEGA_LOCAL_RUNTIME', False)
#: the celery broker name or URL
OMEGA_BROKER = (os.environ.get('OMEGA_BROKER') or
                os.environ.get('RABBITMQ_URL') or
                'amqp://guest@127.0.0.1:5672//')
#: (deprecated) the collection used to store ipython notebooks
OMEGA_NOTEBOOK_COLLECTION = 'ipynb'
#: the celery backend name or URL
OMEGA_RESULT_BACKEND = 'amqp'
#: the celery configurations
OMEGA_CELERY_CONFIG = {
    # FIXME should work with json (the default celery serializer)
    'CELERY_ACCEPT_CONTENT': ['pickle', 'json'],
    'CELERY_TASK_SERIALIZER': 'pickle',
    'CELERY_RESULT_SERIALIZER': 'pickle',
    'CELERY_DEFAULT_QUEUE': os.environ.get('CELERY_Q', 'default'),
    'BROKER_URL': OMEGA_BROKER,
    'BROKER_HEARTBEAT': 0,  # due to https://github.com/celery/celery/issues/4980
    'CELERY_RESULT_BACKEND': OMEGA_RESULT_BACKEND,
    'CELERY_ALWAYS_EAGER': True if OMEGA_LOCAL_RUNTIME else False,
    'CELERYBEAT_SCHEDULE': {
        'execute_scripts': {
            'task': 'omegaml.notebook.tasks.execute_scripts',
            'schedule': 60,
        },
    },
    'BROKER_USE_SSL': OMEGA_USESSL,
}
#: enable cloud worker routing
OMEGA_TASK_ROUTING_ENABLED = os.environ.get('OMEGA_TASK_ROUTING_ENABLED', False)
#: celery task packages
OMEGA_CELERY_IMPORTS = ['omegaml',
                        'omegaml.notebook',
                        'omegaml.backends.package']
#: storage backends
OMEGA_STORE_BACKENDS = {
    'sklearn.joblib': 'omegaml.backends.ScikitLearnBackend',
    'ndarray.bin': 'omegaml.backends.npndarray.NumpyNDArrayBackend',
    'virtualobj.dill': 'omegaml.backends.virtualobj.VirtualObjectBackend',
    'pandas.rawdict': 'omegaml.backends.rawdict.PandasRawDictBackend',
    'python.file': 'omegaml.backends.rawfiles.PythonRawFileBackend',
    'python.package': 'omegaml.backends.package.PythonPackageData',
}
OMEGA_STORE_BACKENDS_TENSORFLOW = {
    'tfkeras.h5': 'omegaml.backends.tensorflow.TensorflowKerasBackend',
    'tfkeras.savedmodel': 'omegaml.backends.tensorflow.TensorflowKerasSavedModelBackend',
    'tf.savedmodel': 'omegaml.backends.tensorflow.TensorflowSavedModelBackend',
    'tfestimator.model': 'omegaml.backends.tensorflow.TFEstimatorModelBackend',
}
OMEGA_STORE_BACKENDS_KERAS = {
    'keras.h5': 'omegaml.backends.keras.KerasBackend',
}
OMEGA_STORE_BACKENDS_SQL = {
    'sqlalchemy.conx': 'omegaml.backends.sqlalchemy.SQLAlchemyBackend',
}
#: supported frameworks
if is_test_run:
    OMEGA_FRAMEWORKS = ('scikit-learn', 'tensorflow', 'keras')
else:
    OMEGA_FRAMEWORKS = os.environ.get('OMEGA_FRAMEWORKS') or ('scikit-learn')

#: storage mixins
OMEGA_STORE_MIXINS = [
    'omegaml.mixins.store.ProjectedMixin',
    'omegaml.mixins.store.virtualobj.VirtualObjectMixin',
    'omegaml.mixins.store.package.PythonPackageMixin',
    'omegaml.mixins.store.promotion.PromotionMixin',
]
#: runtimes mixins
OMEGA_RUNTIME_MIXINS = [
    'omegaml.runtimes.mixins.ModelMixin',
    'omegaml.runtimes.mixins.GridSearchMixin',
]
#: mdataframe mixins
OMEGA_MDF_MIXINS = [
    ('omegaml.mixins.mdf.ApplyMixin', 'MDataFrame,MSeries'),
    ('omegaml.mixins.mdf.FilterOpsMixin', 'MDataFrame,MSeries'),
    ('omegaml.mixins.mdf.apply.ApplyStatistics', 'MDataFrame,MSeries'),
]
#: mdataframe apply context mixins
OMEGA_MDF_APPLY_MIXINS = [
    ('omegaml.mixins.mdf.ApplyArithmetics', 'MDataFrame,MSeries'),
    ('omegaml.mixins.mdf.ApplyDateTime', 'MDataFrame,MSeries'),
    ('omegaml.mixins.mdf.ApplyString', 'MDataFrame,MSeries'),
    ('omegaml.mixins.mdf.ApplyAccumulators', 'MDataFrame,MSeries'),
]
#: user extensions
OMEGA_USER_EXTENSIONS = os.environ.get('OMEGA_USER_EXTENSIONS') or None
#: log dataset
OMEGA_LOG_DATASET = '.omega/logs'
#: OmegaLoggingHandler log format
OMEGA_LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

# =========================================
# ----- DO NOT MODIFY BELOW THIS LINE -----
# =========================================
def update_from_config(vars=globals(), config_file=OMEGA_CONFIG_FILE):
    """
    update omegaml.defaults from configuration file

    :param vars: the variables to update
    :param config_file: the path to config.yml or a file object
    :return:
    """
    # override from configuration file
    userconfig = {}
    if isinstance(config_file, six.string_types):
        if os.path.exists(config_file):
            with open(config_file, 'r') as fin:
                userconfig = yaml.safe_load(fin)
    elif hasattr(config_file, 'read'):
        userconfig = yaml.safe_load(config_file)
    if userconfig:
        for k in [k for k in vars.keys() if k.startswith('OMEGA')]:
            value = userconfig.get(k, None) or vars[k]
            if isinstance(vars[k], dict):
                vars[k].update(value)
            else:
                vars[k] = value
    return vars


def update_from_env(vars=globals()):
    # simple override from env vars
    # -- top-level OMEGA_*
    for k in [k for k in vars.keys() if k.startswith('OMEGA')]:
        vars[k] = os.environ.get(k, None) or vars[k]
    # -- OMEGA_CELERY_CONFIG updates
    for k in [k for k in os.environ.keys() if k.startswith('OMEGA_CELERY')]:
        celery_k = k.replace('OMEGA_', '')
        vars['OMEGA_CELERY_CONFIG'][celery_k] = os.environ[k]
    # -- debug if required
    if '--print-omega-defaults' in sys.argv:
        from pprint import pprint
        vars = {k: v for k, v in six.iteritems(vars) if k.startswith('OMEGA')}
        pprint(vars)
    return vars


def update_from_obj(obj, vars=globals(), attrs=None):
    """
    helper function to update omegaml.defaults from arbitrary module

    :param obj: the source object (must support getattr). Any
       variable starting with OMEGA is set in omegaml.defaults
    """
    for k in [k for k in dir(obj) if k.startswith('OMEGA')]:
        if hasattr(obj, k):
            value = getattr(obj, k)
            if attrs:
                setattr(attrs, k, value)
            else:
                vars[k] = value


def update_from_dict(d, vars=globals(), attrs=None):
    """
    helper function to update omegaml.defaults from arbitrary dictionary

    :param d: the source dict (must support [] lookup). Any
       variable starting with OMEGA is set in omegaml.defaults
    """
    for k, v in six.iteritems(d):
        if k.startswith('OMEGA'):
            if attrs:
                setattr(attrs, k, v)
            else:
                vars[k] = v


# load Enterprise Edition if available
try:
    from omegaee import eedefaults

    update_from_obj(eedefaults, vars=globals())
except Exception as e:
    pass


def locate_config_file(configfile=OMEGA_CONFIG_FILE):
    """
    locate the configuration file, if any

    Will search the following locations for the config file:
        1. current directory
        2. user configuration directory
        3. site configuration directory

    The exact location depends on the platform:
        Linux:
            user = ~/.config/omegaml
            site = /etc/xdg/omegaml
        Windows:
            user = C:\Documents and Settings\<User>\Application Data\omegaml\omegaml
            site = C:\Documents and Settings\All Users\Application Data\omegaml\omegaml
        Mac:
            user = ~/Library/Application Support/omegaml
            site = /Library/Application Support/omegaml

    See the appdirs package for details on the platform specific locations of the
    user and site config dir, https://pypi.org/project/appdirs/

    Args:
        configfile: the default config file name or path

    Returns:
        location of the config file or None if not found
    """
    try:
        from appdirs import user_config_dir, site_config_dir
    except:
        # we don't have appdirs installed, this can happen during setup.py. fake it
        user_config_dir = lambda *args: os.path.expanduser('~/.config/omegaml')
        site_config_dir = lambda *args: '/etc/xdg/omegaml'

    if os.path.exists(configfile):
        return configfile
    appdirs_args = ('omegaml', 'omegaml')
    for cfgdir in (os.getcwd(), user_config_dir(*appdirs_args), site_config_dir(*appdirs_args)):
        cfgfile = os.path.join(cfgdir, os.path.basename(configfile))
        if os.path.exists(cfgfile):
            return cfgfile
    return None


def load_user_extensions(extensions=OMEGA_USER_EXTENSIONS, vars=globals()):
    """
    user extensions are extensions to settings

    Usage:
        in config.yml specify, e.g.

            OMEGA_USER_EXTENSIONS:
                OMEGA_STORE_BACKENDS:
                   KIND: path.to.BackendClass
                EXTENSION_LOADER: path.to.module

        This will extend OMEGA_STORE_BACKENDS and load path.to.module. If the
        EXENSION_LOADER given module has a run(vars) method, it will be called
        using the current defaults variables (dict) as input.

    Args:
        extensions (dict): a list of extensions in the form python.path.to.module or
             <setting_name>: <value>

    Returns:
        None
    """
    for k, v in extensions.items():
        omvar = vars.get(k)
        try:
            if isinstance(omvar, list):
                omvar.append(v)
            elif isinstance(omvar, dict):
                omvar.update(v)
            elif k == 'EXTENSION_LOADER':
                from importlib import import_module
                mod = import_module(v)
                if hasattr(mod, 'run'):
                    mod.run(vars)
            else:
                raise ValueError
        except:
            omvar_type = type(omvar)
            k_type = type(v)
            msg = ('user extensions error: cannot apply {k} to {omvar}, '
                   'expected type {omvar_type} got {k_type}').format(**locals())
            raise ValueError(msg)


# -- test
# this is to avoid using production settings during test
if not is_cli_run and is_test_run:
    OMEGA_MONGO_URL = OMEGA_MONGO_URL.replace('/omega', '/testdb')
    OMEGA_LOCAL_RUNTIME = True
    OMEGA_RESTAPI_URL = ''
    logging.getLogger().setLevel(logging.ERROR)
else:
    # overrides in actual operations
    OMEGA_CONFIG_FILE = locate_config_file()
    update_from_config(globals(), config_file=OMEGA_CONFIG_FILE)
    update_from_env(globals())

    if is_cli_run:
        # be les
        import warnings

        warnings.filterwarnings("ignore", category=FutureWarning)

# load framework-specific backends
# -- note we do this here to ensure this happens after config updates
if 'tensorflow' in OMEGA_FRAMEWORKS and tensorflow_available():
    #: tensorflow backend
    # https://stackoverflow.com/a/38645250
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = os.environ.get('TF_CPP_MIN_LOG_LEVEL') or '3'
    logging.getLogger('tensorflow').setLevel(logging.ERROR)
    OMEGA_STORE_BACKENDS.update(OMEGA_STORE_BACKENDS_TENSORFLOW)
#: keras backend
if 'keras' in OMEGA_FRAMEWORKS and keras_available():
    OMEGA_STORE_BACKENDS.update(OMEGA_STORE_BACKENDS_KERAS)
#: sqlalchemy backend
if module_available('sqlalchemy'):
    OMEGA_STORE_BACKENDS.update(OMEGA_STORE_BACKENDS_SQL)

# load user extensions if any
if OMEGA_USER_EXTENSIONS is not None:
    load_user_extensions(OMEGA_USER_EXTENSIONS)



# source https://djangosnippets.org/snippets/2727/
from requests.auth import AuthBase


class AuthenticationEnv(object):
    @classmethod
    def get_omega_for_task(cls, auth=None):
        from omegaml import Omega
        om = Omega()
        return om

    @classmethod
    def get_omega_from_apikey(cls, *args, **kwargs):
        from omegaml import Omega
        om = Omega()
        return om


class OmegaRestApiAuth(AuthBase):
    """
    Sets the appropriate authentication headers
    for the Omega REST API key authentication.

    Usage:
        auth = OmegaRestApiAuth('jezdez', 
                         '25fdd0d9d210acb78b5b845fe8284a3c93630252')
        response = requests.get('http://api.foo.bar/v1/spam/', auth=auth)
    """

    def __init__(self, username, apikey, qualifier=None):
        self.username = username
        self.apikey = apikey
        self.qualifier = qualifier or 'default'

    @classmethod
    def make_from(cls, om):
        args = getattr(om.runtime, 'auth_tuple', (None, None, 'default'))
        return OmegaRestApiAuth(*args)

    def get_credentials(self):
        return 'ApiKey %s:%s' % (self.username, self.apikey)

    def __call__(self, r):
        r.headers['Authorization'] = self.get_credentials()
        r.headers['Qualifier'] = self.qualifier
        return r

    def __repr__(self):
        return ('OmegaRestApiAuth(username={}, apikey="*****",'
                'qualifier={})').format(self.username, self.qualifier)


class OmegaRuntimeAuthentication:
    """
    The runtimes authentication
    """

    def __init__(self, userid, apikey, qualifier='default'):
        self.userid = userid
        self.apikey = apikey
        self.qualifier = qualifier

    def __repr__(self):
        return ('OmegaRuntimeAuthentication(userid={}, '
                'apikey="*****", qualifier={})').format(self.userid, self.qualifier)


class OmegaSecureAuthenticationEnv(AuthenticationEnv):
    @classmethod
    def get_omega_for_task(cls, auth=None):
        """
        Get Omega instance configured for user in auth

        If auth is passed, a request is made to OMEGA_RESTAPI_URL to
        retrieve the configuration object for this user.

        If auth is the tuple (None, None, 'default') the omegaml module
        is returned, which is configured to the default instance with
        authentication according to the installation. To raise an
        error instead set settings OMEGA_ALLOW_TASK_DEFAULT_AUTH=False

        :param auth: the OmegaRuntimeAuthentication object
        :return: the Omega instance configured for the user
        """
        default_auth = (None, None, 'default')
        is_auth_provided = lambda auth: (auth is not None
                                         and auth != default_auth)
        if is_auth_provided(auth):
            if isinstance(auth, (list, tuple)):
                # we get a serialized tuple, recreate auth object
                # -- this is a hack to easily support python 2/3 client/server mix
                userid, apikey, qualifier = auth
                om = cls.get_omega_from_apikey(userid, apikey, qualifier=qualifier)
            else:
                raise ValueError(
                    'cannot parse authentication as {}'.format(auth))
        elif auth == default_auth:
            # we provide the default implementation as per configuration
            import omegaml
            from omegaml.util import settings
            om = omegaml
            if not settings().OMEGA_ALLOW_TASK_DEFAULT_AUTH:
                raise ValueError(
                    'Default task authentication is not allowed, got {}'.format(auth))
        else:
            raise ValueError(
                'need authentication tupleas (userid, apikey, qualifier), got {}'.format(auth))
        return om

    @classmethod
    def get_omega_from_apikey(cls, *args, **kwargs):
        from omegaml.client.userconf import get_omega_from_apikey
        return get_omega_from_apikey(*args, **kwargs)



import os
import sys
from unittest import TestCase

import numpy as np
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV

from omegaml import Omega
from omegaml.tests.cli.scenarios import CliTestScenarios
from omegaml.tests.util import OmegaTestMixin


class CliRuntimeTests(CliTestScenarios, OmegaTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.om = Omega()
        self.clean()

    def test_cli_runtime_model_fit_predict(self):
        self.make_model('reg')
        reg = self.om.models.get('reg')
        self.make_dataset_from_dataframe('test', N=10)
        # -- fit the model
        self.cli('runtime model reg fit test[x] test[y]')
        self.assertLogContains('info', 'Metadata')
        # -- run a prediction, we get back actual data
        self.cli('runtime model reg predict test[x]', new_start=True)
        self.assertLogContains('info', '[[')
        self.assertIsInstance(self.get_log('info')[0][0], np.ndarray)
        # -- run a prediction, we store the results
        self.cli('runtime model reg predict test[x] --result predictions')
        #    expect metadata since we ask to write data to predictions
        self.assertLogContains('info', 'Metadata')
        df_expect = self.om.datasets.get('test')['y']
        #    check the data has actually been stored as requested
        df_pred = self.om.datasets.get('predictions')
        self.assertIsNotNone(df_pred)
        self.assertEqual(len(df_expect), len(df_pred))

    def test_cli_runtime_model_gridsearch(self):
        X, y = make_classification()
        logreg = LogisticRegression(solver='liblinear')
        self.om.models.put(logreg, 'logreg')
        # perform gridsearch on runtime
        self.om.datasets.put(X, 'testX')
        self.om.datasets.put(y, 'testY')
        self.cli('runtime model logreg gridsearch testX testY --param C=[.1,.5,1.0]')
        model_meta = self.om.models.metadata('logreg')
        self.assertIn('gridsearch', model_meta.attributes)
        gsmodel_name = model_meta.attributes['gridsearch'][0]['gsModel']
        gsmodel = self.om.models.get(gsmodel_name)
        self.assertIsInstance(gsmodel, GridSearchCV)

    def test_cli_runtime_script_run(self):
        pkgpath = self.get_package_path()
        self.cli(f'scripts put {pkgpath} helloworld')
        self.assertLogContains('info', 'Metadata')
        # run without kwargs
        self.cli('runtime script helloworld run', new_start=True)
        self.assertLogContains('info', 'hello from helloworld')
        # with kwargs
        self.cli('runtime script helloworld run foo=bar', new_start=True)
        self.assertLogContains('info', '"foo": "bar"')





class GridSearchMixin(object):
    def gridsearch(self, Xname, Yname, parameters=None, pure_python=False, **kwargs):
        gs_task = self.task('omegaml.tasks.omega_gridsearch')
        Xname = self._ensure_data_is_stored(Xname, prefix='_fitX')
        if Yname is not None:
            Yname = self._ensure_data_is_stored(Yname, prefix='_fitY')
        return gs_task.delay(self.modelname, Xname, Yname, parameters=parameters, **kwargs)



import os

import six
from mongoengine import GridFSProxy

from omegaml.backends import BaseModelBackend
from omegaml.util import temp_filename, remove_temp_filename, ensure_python_array

import numpy as np


class KerasBackend(BaseModelBackend):
    KIND = 'keras.h5'

    @classmethod
    def supports(self, obj, name, **kwargs):
        from keras import Sequential, Model
        return isinstance(obj, (Sequential, Model))

    def _save_model(self, model, fn):
        # override to implement model saving
        model.save(fn)

    def _load_model(self, fn):
        # override to implement model loading
        from keras.engine.saving import load_model
        return load_model(fn)

    def put_model(self, obj, name, attributes=None, **kwargs):
        fn = temp_filename()
        self._save_model(obj, fn)
        with open(fn, mode='rb') as fin:
            fileid = self.model_store.fs.put(
                fin, filename=self.model_store._get_obj_store_key(name, 'h5'))
            gridfile = GridFSProxy(grid_id=fileid,
                                   db_alias='omega',
                                   collection_name=self.model_store.bucket)
        remove_temp_filename(fn)
        return self.model_store._make_metadata(
            name=name,
            prefix=self.model_store.prefix,
            bucket=self.model_store.bucket,
            kind=self.KIND,
            attributes=attributes,
            gridfile=gridfile).save()

    def get_model(self, name, version=-1):
        filename = self.model_store._get_obj_store_key(name, 'h5')
        packagefname = os.path.join(self.model_store.tmppath, name)
        dirname = os.path.dirname(packagefname)
        try:
            os.makedirs(dirname)
        except OSError:
            # OSError is raised if path exists already
            pass
        outf = self.model_store.fs.get_version(filename, version=version)
        with open(packagefname, 'wb') as fout:
            fout.write(outf.read())
        model = self._load_model(packagefname)
        return model

    def fit(self, modelname, Xname, Yname=None, validation_data=None,
            pure_python=True, **kwargs):
        """
        Fit a model

        Args:
            modelname: the name of the model
            Xname: the name of the X dataset
            Yname: the name of the Y dataset
            pure_python: deprecated
            kwargs: any additional kwargs are passed on to model.fit()

        Returns:
            the meta data object of the model
        """
        meta = self.model_store.metadata(modelname)
        model = self.get_model(modelname)
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname) if Yname else None
        keras_kwargs = dict(kwargs)
        if validation_data:
            valX, valY = validation_data
            if isinstance(X, six.string_types):
                valX = self.data_store.get(valX)
            if isinstance(Y, six.string_types):
                valY = self.data_store.get(valY)
            keras_kwargs['validation_data'] = (valX, valY)
        history = model.fit(X, Y, **keras_kwargs)
        meta = self.put_model(model, modelname, attributes={
            'history': serializable_history(history)
        })
        return meta

    def predict(self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        model = self.get_model(modelname)
        X = self.data_store.get(Xname)
        result = model.predict(X)
        if rName:
            meta = self.data_store.put(result, rName)
            result = meta
        return result

    def score(
            self, modelname, Xname, Yname, rName=True, pure_python=True,
            **kwargs):
        model = self.get_model(modelname)
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname)
        score = model.evaluate(X, Y)
        return score


def serializable_history(history):
    # ensure history object is JSON/BSON serializable by converting to default
    return {k: ensure_python_array(v, float)
            for k, v in six.iteritems(history.history)}



class PromotionMixin(object):
    """
    Promote objects from one bucket to another
    """

    def promote(self, name, other):
        """
        Promote object to another bucket.

        This effectively copies the object. If the objects exists in the
        target it will be replaced.

        Args:
            name: The name of the object
            bucket: the bucket to promote to
            other:

        Returns:
            The Metadata of the new object
        """
        if self == other:
            raise ValueError('cannot promote to self')
        # see if the backend supports explicit promotion
        backend = self.get_backend(name)
        if hasattr(backend, 'promote'):
            return backend.promote(name, other)
        # do default promotion
        obj = self.get(name)
        other.drop(name, force=True)
        return other.put(obj, name)



from unittest import TestCase

import numpy as np
from sklearn.linear_model import LinearRegression

from omegaml import Omega
from omegaml.mixins.store.promotion import PromotionMixin
from omegaml.tests.util import OmegaTestMixin


class PromotionMixinTests(OmegaTestMixin, TestCase):
    def setUp(self):
        om = self.om = Omega()
        om.datasets.register_mixin(PromotionMixin)
        om.models.register_mixin(PromotionMixin)
        self.clean()
        self.clean('prod')

    def test_dataset_promotion(self):
        om = self.om
        prod = om['prod']
        om.datasets.put(['foo'], 'foo')
        # ensure dataset is in default bucket, not in prod
        self.assertIn('foo', om.datasets.list())
        self.assertNotIn('foo', prod.datasets.list())
        # promote to prod
        om.datasets.promote('foo', prod.datasets)
        self.assertIn('foo', prod.datasets.list())

    def test_model_promotion(self):
        om = self.om
        prod = om['prod']
        reg = LinearRegression()
        om.models.put(reg, 'mymodel')
        # ensure dataset is in default bucket, not in prod
        self.assertIn('mymodel', om.models.list())
        self.assertNotIn('mymodel', prod.models.list())
        # promote to prod
        om.models.promote('mymodel', prod.models)
        self.assertIn('mymodel', prod.models.list())

    def test_promotion_runtime(self):
        om = self.om
        prod = om['prod']
        reg = LinearRegression()
        om.models.put(reg, 'mymodel')
        # ensure dataset is in default bucket, not in prod
        self.assertIn('mymodel', om.models.list())
        self.assertNotIn('mymodel', prod.models.list())
        # try running on default runtime
        X = np.random.randint(0, 100, (100,)).reshape(-1, 1)
        Y = X * 2
        result = om.runtime.model('mymodel').fit(X, Y).get()
        assert 'Metadata' in result
        # try running on prod runtime -- should not work
        with self.assertRaises(AssertionError):
            prod.runtime.model('mymodel').fit(X, Y).get()
        # promote to prod
        om.models.promote('mymodel', prod.models)
        # try running on prod runtime -- now should work
        result = prod.runtime.model('mymodel').fit(X, Y).get()
        assert 'Metadata' in result
        self.assertIn('mymodel', prod.models.list())

    def test_promotion_deferred(self):
        # ensure we always import injected deferred
        from omegaml import _omega
        om = _omega.OmegaDeferredInstance()
        prod = om['prod']
        reg = LinearRegression()
        om.models.put(reg, 'mymodel')
        # ensure dataset is in default bucket, not in prod
        self.assertIn('mymodel', om.models.list())
        self.assertNotIn('mymodel', prod.models.list())
        # try running on default runtime
        X = np.random.randint(0, 100, (100,)).reshape(-1, 1)
        Y = X * 2
        result = om.runtime.model('mymodel').fit(X, Y).get()
        assert 'Metadata' in result
        # try running on prod runtime -- should not work
        with self.assertRaises(AssertionError):
            prod.runtime.model('mymodel').fit(X, Y).get()
        # promote to prod
        om.models.promote('mymodel', prod.models)
        # try running on prod runtime -- now should work
        result = prod.runtime.model('mymodel').fit(X, Y).get()
        assert 'Metadata' in result
        self.assertIn('mymodel', prod.models.list())

    def test_promotion_to_self_fails(self):
        om = self.om
        reg = LinearRegression()
        om.models.put(reg, 'mymodel')
        with self.assertRaises(ValueError):
            om.models.promote('mymodel', om.models)
        om.datasets.put(['foo'], 'foo')
        with self.assertRaises(ValueError):
            om.datasets.promote('foo', om.datasets)

    def test_promotion_to_other_db_works(self):
        om = self.om
        other = Omega(mongo_url=om.mongo_url + '_test')
        reg = LinearRegression()
        om.models.put(reg, 'mymodel')
        om.models.promote('mymodel', other.models)
        om.datasets.put(['foo'], 'foo')
        om.datasets.promote('foo', other.datasets)




import os

from behave import when, then

from omegaml.tests.features.util import Notebook


@when('we upload the {nbname} notebook')
def uploadtutorial(ctx, nbname):
    # uploading does not work because a native dialogue is opened
    # br.find_by_css('input.fileinput')
    # br.find_by_css('input.fileinput').click() # opens a native file dialoge
    from nbformat import read as nbread
    br = ctx.browser
    om = ctx.feature.om
    # upload directly
    nbfname = os.path.join(ctx.nbfiles, '{nbname}.ipynb'.format(**locals()))
    nbcells = nbread(nbfname, as_version=4)
    om.jobs.put(nbcells, nbname)
    # now run the notebook
    br.visit(ctx.feature.jynb_url)
    assert br.is_text_present(nbname)


@when('we run the notebook {nbname}')
def runnotebook(ctx, nbname):
    br = ctx.browser
    om = ctx.feature.om
    br.visit(ctx.feature.jynb_url)
    nb = Notebook(br)
    # FIXME sometimes it takes long for the nb to appear. why?
    nb.open_notebook(nbname, retry=10)
    nb.run_all_cells(wait=True)
    nb.save_notebook()
    assert not br.is_text_present('Error')
    assert not br.is_text_present('Exception')
    assert not br.is_text_present('failed')
    assert not br.is_text_present('MissingSchema')
    assert not br.is_text_present('error')


@then('model {model_name} exists')
def checkmodel(ctx, model_name):
    br = ctx.browser
    om = ctx.feature.om
    assert model_name in om.models.list()

@then('dataset {dataset_name} exists')
def checkdataset(ctx, dataset_name):
    br = ctx.browser
    om = ctx.feature.om
    assert dataset_name in om.datasets.list()



import json

import datetime
import os
import sys
from shutil import rmtree
from unittest import TestCase

from omegaml import Omega
from omegaml.backends.package import PythonPackageData
from omegaml.backends.package.packager import build_sdist
from omegaml.util import settings, delete_database, mkdirs


class PythonPackageDataTests(TestCase):
    _multiprocess_can_split_ = False

    def setUp(self):
        self.basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        self.defaults = settings()
        OMEGA_STORE_BACKENDS = self.defaults.OMEGA_STORE_BACKENDS
        OMEGA_STORE_MIXINS = self.defaults.OMEGA_STORE_MIXINS
        self.backend = backend = 'omegaml.backends.package.PythonPackageData'
        self.mixin = mixin = 'omegaml.mixins.store.package.PythonPackageMixin'
        if PythonPackageData.KIND not in OMEGA_STORE_BACKENDS:
            OMEGA_STORE_BACKENDS[PythonPackageData.KIND] = backend
        if mixin not in OMEGA_STORE_MIXINS:
            OMEGA_STORE_MIXINS.append(mixin)
        self.om = Omega()
        delete_database()
        self.pkgsdir = self.om.scripts.get_backend_bykind(PythonPackageData.KIND).packages_path
        self.distdir = os.path.join(self.basepath,  'demo', 'helloworld', 'dist')
        rmtree(self.distdir, ignore_errors=True)
        mkdirs(self.distdir)
        mkdirs(self.pkgsdir)

    def tearDown(self):
        rmtree(self.pkgsdir, ignore_errors=True)
        rmtree(self.distdir)

    def test_build_sdist(self):
        pkgpath = os.path.abspath(os.path.join(self.basepath, 'demo', 'helloworld'))
        distdir = os.path.join(pkgpath, 'dist')
        sdist = build_sdist(pkgpath, self.distdir)
        version = sdist.metadata.version
        pkgname = sdist.metadata.name
        pkgdist = os.path.join(distdir, '{pkgname}-{version}.tar.gz'.format(**locals()))
        self.assertTrue(os.path.exists(pkgdist))

    def test_build_put(self):
        om = self.om
        pkgpath = os.path.abspath(os.path.join(self.basepath, 'demo', 'helloworld'))
        pkgsrc = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkgsrc, 'helloworld')
        metas = om.scripts.list(raw=True)
        self.assertEqual(metas[0].kind, 'python.package')

    def test_build_get(self):
        om = self.om
        pkgpath = os.path.abspath(os.path.join(self.basepath, 'demo', 'helloworld'))
        pkgsrc = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkgsrc, 'helloworld')
        # load and install
        # -- since this is loading from gridfs first and installing the package using pip
        #    this will be orders of magnitues slower than just getting the package when
        #    already installed the second time around (see second step & test)
        dtstart = datetime.datetime.now()
        mod1 = om.scripts.get('helloworld')
        dtend = datetime.datetime.now()
        loadtime1 = dtend - dtstart
        # see if we can just get from the existing local installation
        # -- this will just load the module from disk
        dtstart = datetime.datetime.now()
        mod2 = om.scripts.get('helloworld')
        dtend = datetime.datetime.now()
        loadtime2 = dtend - dtstart
        # assert that loading from the local module from disk is faster than loading from gridfs
        self.assertLess(loadtime2, loadtime1)
        self.assertEqual(mod2.__file__, mod2.__file__)

    def test_install(self):
        om = self.om
        pkgpath = os.path.abspath(os.path.join(self.basepath, 'demo', 'helloworld'))
        pkgsrc = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkgsrc, 'helloworld')
        om.scripts.install()
        self.assertIn('helloworld', sys.modules)

    def test_runtime(self):
        om = self.om
        pkgpath = os.path.abspath(os.path.join(self.basepath, 'demo', 'helloworld'))
        pkgsrc = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkgsrc, 'helloworld')
        print("***omega test_runtime (om, om.runtime)", om, om.runtime)
        result = om.runtime.script('helloworld').run(text='foo')
        data = json.loads(result.get())
        self.assertIn('runtimes', data)
        expected = ['hello from helloworld', {'text': 'foo', 'pure_python': False}]
        self.assertEqual(data['result'], expected)

    def test_sysargv_stability(self):
        """
        test scripts.put does not change sys.argv
        """
        om = self.om
        orig_sysargv = ' '.join(sys.argv)
        pkgpath = os.path.abspath(os.path.join(self.basepath, 'demo', 'helloworld'))
        pkgsrc = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkgsrc, 'helloworld')
        new_sysargv = ' '.join(sys.argv)
        self.assertEqual(new_sysargv, orig_sysargv)







import imghdr

import pandas as pd

from omegaml.client.docoptparser import CommandBase
from omegaml.client.util import get_omega


class DatasetsCommandBase(CommandBase):
    """
    Usage:
      om datasets list [<pattern>] [--raw] [-E|--regexp] [options]
      om datasets put <path> <name> [--replace] [--csv=<param=value>]... [options]
      om datasets get <name> <path> [--csv <param>=<value>]... [options]
      om datasets drop <name> [--force] [options]
      om datasets metadata <name> [options]

    Options:
      --raw   return metadata

    Description:
         For csv files, put and get accept the --csv option multiple times.
         The <param>=<value> pairs will be used as kwargs to pd.read_csv (on put)
         and df.to_csv methods (on get)
    """
    command = 'datasets'

    def list(self):
        om = get_omega(self.args)
        raw = self.args.get('--raw', False)
        regexp = self.args.get('--regexp') or self.args.get('-E')
        pattern = self.args.get('<pattern>')
        kwargs = dict(regexp=pattern) if regexp else dict(pattern=pattern)
        entries = om.datasets.list(raw=raw, **kwargs)
        self.logger.info(entries)

    def put(self):
        om = get_omega(self.args)
        local = self.args['<path>']
        name = self.args['<name>']
        replace = self.args['--replace']
        csvkwargs = self.parse_kwargs('--csv')
        # TODO introduce a puggable filetype processing backend to do this
        if local.endswith('.csv'):
            # csv formats
            import pandas as pd
            data = pd.read_csv(local, **csvkwargs)
            meta = om.datasets.put(data, name, append=not replace)
        elif imghdr.what(local):
            # images
            from imageio import imread
            img = imread(local)
            meta = om.datasets.put(img, name)
        else:
            meta = self.logger.info(om.datasets.put(local, name, append=not replace))
        self.logger.info(meta)

    def load(self):
        return self.put()

    def export(self):
        return self.get()

    def get(self):
        om = get_omega(self.args)
        local = self.args['<path>']
        name = self.args['<name>']
        obj = om.datasets.get(name)
        csvkwargs = self.parse_kwargs('--csv', index=False)
        if isinstance(obj, pd.DataFrame):
            obj.to_csv(local, **csvkwargs)
        elif hasattr(obj, 'read'):
            with open(local, 'wb') as fout:
                while True:
                    data = obj.read(1024 * 10)
                    if not data:
                        break
                    fout.write(data)
        self.logger.debug(local)

    def drop(self):
        om = get_omega(self.args)
        name = self.args['<name>']
        force = self.args['--force']
        om.datasets.drop(name, force=force)

    def metadata(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        self.logger.info(om.datasets.metadata(name).to_json())

    def plugins(self):
        om = get_omega(self.args)
        for kind, plugincls in om.defaults.OMEGA_STORE_BACKENDS.items():
            self.logger.info(kind, plugincls.__doc__)



from importlib import import_module

from omegaml.client.docoptparser import CommandBase
from omegaml.client.util import get_omega


class ModelsCommandBase(CommandBase):
    """
    Usage:
      om models put <module.callable> <name>
      om models list [<pattern>] [--raw] [-E|--regexp] [options]
      om models metadata <name>

    Description:
        Great stuff with models
    """
    command = 'models'

    def list(self):
        om = get_omega(self.args)
        pattern = self.args.get('<pattern>')
        regexp = self.args.get('--regexp') or self.args.get('-E')
        raw = self.args.get('--raw')
        kwargs = dict(regexp=pattern) if regexp else dict(pattern=pattern)
        self.logger.info(om.models.list(raw=raw, **kwargs))

    def put(self):
        om = get_omega(self.args)
        modname = self.args.get('<module.callable>')
        name = self.args.get('<name>')
        modname, modelfn = modname.rsplit('.', maxsplit=1)
        try:
            mod = import_module(modname)
        except:
            raise
        modelfn = getattr(mod, modelfn)
        model = modelfn()
        self.logger.info(om.models.put(model, name))

    def metadata(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        self.logger.info(om.models.metadata(name).to_json())







from unittest.case import TestCase

import numpy as np
from sklearn.datasets import make_regression
from sklearn.linear_model.stochastic_gradient import SGDRegressor
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from omegaml.sklext import OnlinePipeline


class OnlinePipelineTests(TestCase):
    def test_partial_fit(self):
        # define an online pipeline
        piple = OnlinePipeline([
            ('scale', StandardScaler()),
            ('clf', SGDRegressor(random_state=5, shuffle=False, verbose=True, max_iter=10)),
        ])
        # define an offline pipelines
        pipl = Pipeline([
            ('scale', StandardScaler()),
            ('clf', SGDRegressor(random_state=5, shuffle=False, verbose=True, max_iter=20)),
        ])
        # generate some data
        X, y = make_regression(100, 100, random_state=42)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=.33, random_state=42)
        # fit, predict in an online manner
        for i in range(100):
            piple.partial_fit(X_train[0:30], y_train[0:30])
            piple.partial_fit(X_train[30:], y_train[30:])
        ye = piple.predict(X_test)
        # fit, predict in offline manner
        pipl.fit(X_train, y_train)
        yh = pipl.predict(X_test)
        # compare results
        r2_offline = r2_score(y_test, yh)
        r2_online = r2_score(y_test, ye)
        # use a relatively high tolerance due to differences in results of offline and online
        np.testing.assert_allclose(r2_offline, r2_online, atol=0.1)




class BaseDataBackend(object):

    """
    OmegaML BaseDataBackend to be subclassed by other arbitrary backends

    This provides the abstract interface for any data backend to be implemented
    """
    def __init__(self, model_store=None, data_store=None, **kwargs):
        assert model_store, "Need a model store"
        assert data_store, "Need a data store"
        self.model_store = model_store
        self.data_store = data_store

    @classmethod
    def supports(self, obj, name, **kwargs):
        """
        test if this backend supports this obj
        """
        return False

    def put(self, obj, name, attributes=None, **kwargs):
        """
        put an obj 

        :param obj: the object to store (object)
        :param name: the name of the object (str)
        :param attributes: the attributes dict (dict, optional)
        :param kwargs: other kwargs to be passed to the Metadata object
        :return: the Metadata object
        """
        raise NotImplementedError

    def get(self, name, version=-1, force_python=False, lazy=False, **kwargs):
        """
        get an obj 

        :param name: the name of the object (str)
        :return: the object as it was originally stored
        """
        raise NotImplementedError

    def getl(self, *args, **kwargs):
        """
        get an lazy implementation to access the obj

        A lazy implementation is a proxy to the object that can be 
        evaluated using the :code:`.value` property. The proxy should
        ensure that any operations applied on the object are delayed until
        the .value property is accessed. Typically this is to ensure that
        the actual computation is executed on the cluster, not on the local
        machine. 

        :param name: the name of the object (str)
        :return: the proxy to the object as it was originally stored
        """
        return self.get(*args, lazy=True, **kwargs)



from pandas.io.json import json_normalize
from pymongo.collection import Collection

from omegaml.backends.basedata import BaseDataBackend
from omegaml.mdataframe import MDataFrame


class PandasRawDictBackend(BaseDataBackend):
    """
    OmegaStore backend to support arbitrary collections

    Usage:
        # store any collection as part of metadata
        coll = db['some_collection']
        om.datasets.put(coll, 'foo')
        => Metadata(name='foo', collection='some_collection', ...)
        # parse the collection using pandas.io.json_normalize
        df = om.datasets.get('foo')
        # use an alternate parser that accepts dict|list(dict)
        df = om.datasets.get('foo', parser=some_fn)
        # get a MDataFrame
        om.datasets.getl('foo')
        # preserve all document keys, including _id
        om.datasets.getl('foo', raw=True)
    """
    KIND = 'pandas.rawdict'

    @classmethod
    def supports(self, obj, name, as_raw=None, **kwargs):
        return (as_raw and isinstance(obj, dict)) or isinstance(obj, Collection)

    def get(self, name, version=-1, lazy=False, raw=False, parser=None, **kwargs):
        collection = self.data_store.collection(name)
        # json_normalize needs a list of dicts to work, not a generator
        json_normalizer = lambda v: json_normalize([r for r in v])
        parser = parser or json_normalizer
        mdf = MDataFrame(collection, query=kwargs, parser=parser, raw=raw, **kwargs)
        return mdf if lazy else mdf.value

    def put(self, obj, name, attributes=None, **kwargs):
        if isinstance(obj, dict):
            # actual data, just insert
            collection = self.data_store.collection(name)
            collection.insert(obj)
        else:
            # already a collection, import it to metadata
            collection = obj
        meta = self.data_store._make_metadata(name,
                                              kind=self.KIND,
                                              collection=collection.name,
                                              attributes=attributes,
                                              **kwargs)
        return meta.save()



import logging
import os
import sys
import threading

import callable_pip as cpip
from tee import tee

sync_lock = threading.Lock()

logger = logging.getLogger(__name__)

def build_sdist(src, distdir):
    from distutils.core import run_setup
    # save argv. distutils run_setup changes argv and fails at restoring properly
    save_argv = list(sys.argv)
    # block any other thread from executing since we're changing cwd and sys.argv
    with sync_lock, tee.StdoutTee('build_sdist.out', 'w', 2), \
         tee.StderrTee('build_sdist.err', 'w', 2):
        # make sure no other processing is executing while we change the working directory for setup.py
        cwd = os.getcwd()
        sys.stdout.errors = None
        sys.stderr.errors = None
        try:
            os.chdir(src)
            setup_path = os.path.join(src, 'setup.py')
            sdist = run_setup(setup_path, script_args=['sdist', '--dist-dir', distdir])
        finally:
            # restore cwd
            os.chdir(cwd)
            # restore sys argv
            for i ,v in enumerate(save_argv):
                sys.argv[i] = v
            if len(sys.argv) > len(save_argv):
                del sys.argv[len(save_argv):]
    return sdist


def install_package(src, dst):
    cpip.main('install',
              src,
              '--force-reinstall',
              '--upgrade',
              '--target', dst)


def load_from_path(name, path):
    import importlib
    sys.path.insert(0, path)
    try:
        if name in sys.modules:
            del sys.modules[name]
        if name in globals():
            del globals()[name]
        mod = importlib.import_module(name)
    finally:
        sys.path.remove(path)
    return mod


def install_and_import(pkgfilename, package, installdir):
    install_package(pkgfilename, installdir)
    mod = load_from_path(package, installdir)
    return mod


if __name__ == '__main__':
    basepath = os.path.join(os.path.dirname(__file__), '..')
    pkgpath = os.path.join(basepath, 'demo/helloworld')
    build_sdist(pkgpath)

    pkgfilename = os.path.join(basepath, 'demo/helloworld/dist/helloworld-1.0.tar.gz')
    installdir = '/tmp/xyz'
    install_and_import(pkgfilename, 'helloworld', installdir)

    import helloworld

    helloworld.hello()



from getpass import getuser
from logging import warning

import logging
import os
import pandas as pd
import six

from omegaml.backends.basedata import BaseDataBackend

try:
    import snowflake

    sql_logger = logging.getLogger('snowflake')
    sql_logger.setLevel('CRITICAL')
except:
    pass

CNX_CACHE = {}
ALWAYS_CACHE = False


class SQLAlchemyBackend(BaseDataBackend):
    """
    sqlalchemy plugin for omegaml

    Usage:
        # define your sqlalchemy connection
        sqlalchemy_constr = f'sqlalchemy://{user}:{password}@{account}/'

        Store in any of three ways:

        # -- just the connection
        om.datasets.put(sqlalchemy_constr, 'mysqlalchemy')
        om.datasets.get('mysqlalchemy')
        => the sql connection object

        # -- store connection with a predefined sql
        om.datasets.put(sqlalchemy_constr, 'mysqlalchemy', sql='select ....')
        om.datasets.get('mysqlalchemy')
        => will return a pandas dataframe. specify chunksize to return an interable of dataframes

        # -- predefined sqls can contain variables to be resolved at access time
        #    if you miss to specify required variables in sqlvars, a KeyError is raised
        om.datasets.put(sqlaclhemy_constr, 'myview', sql='select ... from col="{var}"')
        om.datasets.get('mysqlalchemy', sqlvars=dict(var="value"))

        Copy to an omega dataset from the connection

        # -- copy the result of the sqlalchemy query to omegaml
        om.datasets.put(sqlalchemy_constr, 'mysqlalchemy', sql='select ...', copy=True)
        om.datasets.get('mysqlalchemy')
        => will return a pandas dataframe (without executing any additional queries)
        => can also use with om.datasets.getl('mysqlalchemy') to return a MDataFrame

        Insert data via the connection

        # -- store data back through the connection
        om.datasets.put(sqlalchemy_constr, 'mysqlalchemy')
        om.datasets.put(df, 'mysqlalchemy', table='SOMETABLE')

        Connection strings can contain variables, e.g. userid and password.
        By default variables are resolved from the os environment. Can also
        specify using any dict.

        # -- use connection string with variables
        sqlalchemy_constr = 'sqlite:///{dbname}.db'
        om.datasets.put(sqlalchemy_constr, 'userdb')
        om.datasets.get('userdb', secrets=dict(dbname='chuckdb'))

        # -- alternatively, create a vault dataset:
        secrets = dict(userid='chuck', dbname='chuckdb')
        om.datasets.put(secrets, '_omega/vault')
        om.datasets.get('userdb')

        the '_omega/vault' dataset will be queried using the current userid as
        the secret name,ad the dbname retrieved from the document. This is
        experimental and the vault is not encrypted.

    Advanced:

        om.datasets.put() supports the following additional keyword arguments

        chunksize=int   specify the number of rows to read from sqlalchemy in one chunk.
                        defaults to 10000

        parse_dates=['col', ...] list of column names to parse for date, time or datetime.
                        see pd.read_sql for details

        transform=callable  a callable, is passed the DataFrame of each chunk before it
                        is inserted into the database. use to provide custom transformations.
                        only works on copy=True

        as well as other kwargs supported by pd.read_sql

    """
    KIND = 'sqlalchemy.conx'

    @classmethod
    def supports(cls, obj, name, insert=False, data_store=None, model_store=None, *args, **kwargs):
        valid = cls._is_valid_url(cls, obj)
        support_via = cls._supports_via(cls, data_store, name, obj)
        return valid or support_via

    def drop(self, name, **kwargs):
        if name in CNX_CACHE:
            del CNX_CACHE[name]
        return super().drop(name, **kwargs)

    def get(self, name, sql=None, chunksize=None, raw=False, sqlvars=None,
            secrets=None, index=True, keep=False, lazy=False, *args, **kwargs):
        """
        retrieve connection or query data from connection

        Args:
            name (str): the name of the connection
            secrets (dict): dict to resolve variables in the connection string
            keep (bool): if True connection is kept open.

        Query data, specify sql='select ...':
            sql (str): the sql query, defaults to the query specific on .put()
            chunksize (int): the number of records for each chunk, if
               specified returns an iterator
            sqlvars (dict): optional, if specified will be used to format sql

        Get connection:
            raw (bool): the raw sql alchemy connection

        To reuse connections:
            specify keep=True. Note this is potentially unsafe in a multi-user
            environment where connection strings contain user-specific secrets.
            If you want to always keep connections open, specify
            om.datasets.defaults.SQLALCHEMY_ALWAYS_CACHE=True

        Returns:
            connection or pd.DataFrame
        """
        meta = self.data_store.metadata(name)
        connection_str = meta.kind_meta.get('sqlalchemy_connection')
        sql = sql or meta.kind_meta.get('sql')
        chunksize = chunksize or meta.kind_meta.get('chunksize')
        keep = getattr(self.data_store.defaults, 'SQLALCHEMY_ALWAYS_CACHE',
                       ALWAYS_CACHE) or keep
        if connection_str:
            secrets = self._get_secrets(meta, secrets)
            connection = self._get_connection(name, connection_str, secrets=secrets, keep=keep)
        else:
            raise ValueError('no connection string')
        if not raw and sql:
            index_cols = _meta_to_indexcols(meta) if index else kwargs.get('index_col')
            try:
                sql = sql.format(**(sqlvars or {}))
            except KeyError as e:
                raise KeyError('{e}, specify sqlvars= to build query >{sql}<'.format(**locals()))
            kwargs = meta.kind_meta.get('kwargs') or {}
            kwargs.update(kwargs)
            if not lazy:
                result = pd.read_sql(sql, connection, chunksize=chunksize, index_col=index_cols, **kwargs)
            else:
                # lazy returns a cursor
                result = connection.execute(sql)
            if not keep:
                connection.close()
            return result
        return connection

    def put(self, obj, name, sql=None, copy=False, append=True, chunksize=None,
            transform=None, table=None, attributes=None, insert=False,
            secrets=None, *args, **kwargs):
        """
        store sqlalchemy connection or data into an existing connection

        Args:
            obj (str|pd.DataFrame): the sqlalchemy connection string or a dataframe object
            name (str): the name of the object
            table (str): optional, if specified is stored along connection
            sql (str): optional, if specified is stored along connection
            copy (bool): optional, if True the connection is queried using sql and the resulting data is stored instead,
                see Copying
            attributes (dict): optional, set or update metadata.attributes

        Copying data, specify copy=True:
            sql (str): sql to query
            append (bool): if True the data is appended if exists already
            chunksize (int): number of records to use
            transform (callable): passed as DataFrame.to_sql(method=)

        Inserting via connection, specify insert=True:
            insert (bool): specify True to insert via connection
            table (str): the table name to use for inserting data
            append (bool): if False will replace any existing table

        Returns:
            metadata
        """
        if not insert and self._is_valid_url(obj):
            # store a connection object
            url = obj
            cnx_name = name if not copy else '_cnx_{}'.format(name)
            metadata = self._put_as_connection(url, cnx_name, sql=sql, chunksize=chunksize,
                                               table=table, attributes=attributes, **kwargs)
            if copy:
                secrets = self._get_secrets(metadata, secrets)
                self._put_as_data(url, name, cnx_name,
                                  sql=sql, chunksize=chunksize,
                                  append=append, transform=None,
                                  secrets=secrets,
                                  **kwargs)
        elif self.data_store.metadata(name) is not None:
            metadata = self._put_via(obj, name, append=append, table=table, chunksize=chunksize,
                                     transform=transform, **kwargs)
        else:
            raise ValueError('type {} is not supported by {}'.format(type(obj), self.KIND))
        metadata.attributes.update(attributes) if attributes else None
        return metadata.save()

    def _put_via(self, obj, name, append=True, table=None, chunksize=None, transform=None,
                 index_columns=None, index=True, **kwargs):
        # write data back through the connection
        # -- ensure we have a valid object
        if not hasattr(obj, 'to_sql'):
            warning('obj.to_sql() does not exist, trying pd.DataFrame(obj)')
            obj = pd.DataFrame(obj)
        # -- get the connection
        connection = self.get(name, raw=True)
        metadata = self.data_store.metadata(name)
        table = table or metadata.kind_meta['table'] or name
        if isinstance(obj, pd.DataFrame) and index:
            index_cols = _dataframe_to_indexcols(obj, metadata, index_columns=index_columns)
        else:
            index_cols = index_columns
        metadata.kind_meta['index_columns'] = index_cols
        exists_action = 'append' if append else 'replace'
        transform = transform
        self._chunked_to_sql(obj, table, connection, chunksize=chunksize, method=transform,
                             if_exists=exists_action, index=index, index_label=index_cols, **kwargs)
        connection.close()
        return metadata

    def _put_as_data(self, url, name, cnx_name, sql=None, chunksize=None, append=True,
                     transform=None, **kwargs):
        # use the url to query the connection and store resulting data instead
        if not sql:
            raise ValueError('a valid SQL statement is required with copy=True')
        metadata = self.copy_from_sql(sql, url, name, chunksize=chunksize,
                                      append=append, transform=transform,
                                      **kwargs)
        metadata.attributes['created_from'] = cnx_name
        return metadata

    def _put_as_connection(self, url, name, sql=None, chunksize=None, attributes=None,
                           table=None, index_columns=None, secrets=None, **kwargs):
        kind_meta = {
            'sqlalchemy_connection': str(url),
            'sql': sql,
            'chunksize': chunksize,
            'table': table,
            'index_columns': index_columns,
            'kwargs': kwargs,
        }
        if secrets is True or (secrets is None and '{' in url and '}' in url):
            kind_meta['secrets'] = {
                'dsname': '_omega/vault',
                'query': {
                    'data_userid': '{user}'
                }
            }
        else:
            kind_meta['secrets'] = secrets
        metadata = self.data_store.metadata(name)
        if metadata is not None:
            metadata.kind_meta.update(kind_meta)
        else:
            metadata = self.data_store.make_metadata(name, self.KIND,
                                                     kind_meta=kind_meta,
                                                     attributes=attributes)
        return metadata.save()

    def _get_connection(self, name, connection_str, secrets=None, keep=False):
        from sqlalchemy import create_engine

        connection = None
        try:
            connection_str = connection_str.format(**(secrets or {}))
            engine = create_engine(connection_str, echo=False)
            connection = CNX_CACHE.get(name) or engine.connect()
        except KeyError as e:
            msg = ('{e}, ensure secrets are specified for connection '
                   '>{connection_str}<'.format(**locals()))
            raise KeyError(msg)
        except Exception as e:
            if connection is not None:
                connection.close()
            raise
        else:
            if keep:
                CNX_CACHE[name] = connection
            else:
                if name in CNX_CACHE:
                    del CNX_CACHE[name]
        return connection

    def copy_from_sql(self, sql, connstr, name, chunksize=10000,
                      append=False, transform=None, secrets=None, **kwargs):
        connection = self._get_connection(name, connstr, secrets=secrets)
        chunksize = chunksize or 10000  # avoid None
        pditer = pd.read_sql(sql, connection, chunksize=chunksize, **kwargs)
        try:
            import tqdm
        except:
            meta = self._chunked_insert(pditer, name, append=append,
                                        transform=transform)
        else:
            with tqdm.tqdm(unit='rows') as pbar:
                meta = self._chunked_insert(pditer, name, append=append,
                                            transform=transform, pbar=pbar)
        connection.close()
        return meta

    def _chunked_to_sql(self, df, table, connection, if_exists='append', chunksize=None, pbar=True, **kwargs):
        # insert large df in chunks and with a progress bar
        # from https://stackoverflow.com/a/39495229
        chunksize = chunksize if chunksize is not None else 10000
        def chunker(seq, size):
            return (seq[pos:pos + size] for pos in range(0, len(seq), size))

        def to_sql(df, table, connection, pbar=False):
            for i, cdf in enumerate(chunker(df, chunksize)):
                exists_action = if_exists if i == 0 else "append"
                cdf.to_sql(table, con=connection, if_exists=exists_action, **kwargs)
                if pbar:
                    pbar.update(len(cdf))
                else:
                    print("writing chunk {}".format(i))

        try:
            import tqdm
            if pbar is False:
                pbar = None
        except Exception as e:
            to_sql(df, table, connection, pbar=pbar)
        else:
            with tqdm.tqdm(total=len(df), unit='rows') as pbar:
                to_sql(df, table, connection, pbar=pbar)

    def _chunked_insert(self, pditer, name, append=True, transform=None, pbar=None):
        # insert into om dataset
        for i, df in enumerate(pditer):
            if pbar is not None:
                pbar.update(len(df))
            should_append = (i > 0) or append
            if transform:
                df = transform(df)
            try:
                meta = self.data_store.put(df, name, append=should_append)
            except Exception as e:
                rows = df.iloc[0:10].to_dict()
                raise ValueError("{e}: {rows}".format(**locals()))
        return meta

    def _is_valid_url(self, url):
        # enable subclass override
        return _is_valid_url(url)

    def _supports_via(self, data_store, name, obj):
        obj_ok = isinstance(obj, (pd.Series, pd.DataFrame, dict))
        if obj_ok and data_store:
            meta = data_store.metadata(name)
            same_kind = meta and meta.kind == self.KIND
            return obj_ok and same_kind
        return False

    def _get_secrets(self, meta, secrets):
        secrets_specs = meta.kind_meta.get('secrets')
        if not secrets and secrets_specs:
            dsname = secrets_specs['dsname']
            query = secrets_specs['query']
            # -- format query values
            query = _format_dict(query, replace=('_', '.'), **os.environ, user=getuser())
            # -- run query
            secrets = self.data_store.get(dsname, filter=query)
            secrets = secrets[0] if isinstance(secrets, list) and len(secrets) == 1 else {}
            secrets.update(os.environ)
        # -- format secrets
        if secrets:
            secrets = _format_dict(secrets, **os.environ, user=getuser())
        return secrets


def _is_valid_url(url):
    # check if we have a valid url with a registered backend
    import sqlalchemy

    try:
        url = sqlalchemy.engine.url.make_url(url)
        drivername = url.drivername.split('+')[0] # e.g. mssql+pyodbc => mssql
        valid = url.drivername in sqlalchemy.dialects.__all__
        valid |= sqlalchemy.dialects.registry.load(drivername) is not None
    except:
        valid = False
    return valid


def _dataframe_to_indexcols(df, metadata, index_columns=None):
    # from a dataframe get index column names
    # works like pd.DataFrame.to_sql except for creating default index_i cols
    # for any missing (None) index labels in a MultiIndex.
    index_cols = metadata.kind_meta.get('index_columns') or index_columns or list(df.index.names)
    multi = isinstance(df.index, pd.MultiIndex)
    if index_cols is not None:
        for i, col in enumerate(index_cols):
            if col is None:
                index_cols[i] = 'index' if not multi else 'index_{}'.format(i)
    return index_cols


def _meta_to_indexcols(meta):
    index_cols = meta.kind_meta.get('index_columns')
    multi = isinstance(index_cols, (list, tuple)) and len(index_cols) > 1
    if index_cols is not None and not isinstance(index_cols, six.string_types):
        for i, col in enumerate(index_cols):
            if col is None:
                index_cols[i] = 'index' if not multi else 'index_{}'.format(i)
    return index_cols


def _format_dict(d, replace=None, **kwargs):
    for k, v in d.items():
        if replace:
            del d[k]
            k = k.replace(*replace) if replace else k
        d[k] = v.format(**kwargs)
    return d


def load_sql(om=None, kind=SQLAlchemyBackend.KIND):
    """
    load ipython sql magic, loading all sql alchemy connections

    Usage:
        !pip install ipython-sql

        # prepare some connection, insert some data
        df = pd.DataFrame(...)
        om.datasets.put('sqlite:///test.db', 'testdb')
        om.datasets.put(df, 'testdb', table='foobar', insert=True)

        from omegaml.backends.sqlalchemy import load_sql
        load_sql()

        # list registered connections
        %sql
            omsql://testdb

        # run queries
        %sql omsql://testdb select * from foobar

    See Also:
        https://pypi.org/project/ipython-sql/

    Args:
        om (Omega): optional, specify omega instance, defaults to om.setup()
        kind (str): the backend's kind, used to find connections to register
           with the sql magic

    Returns:
        None
    """
    from unittest.mock import MagicMock
    from IPython import get_ipython
    import omegaml as om
    from sql.connection import Connection

    class ConnectionShim:
        # this is required to trick sql magic into accepting existing connection objects
        # (by default sql magic expects connection strings, not connection objects)
        def __init__(self, url, conn):
            self.session = conn
            self.metadata = MagicMock()
            self.metadata.bind.url = url
            self.dialect = getattr(conn, 'dialect', 'omsql')

    # load sql magic
    ipython = get_ipython()
    ipython.magic('load_ext sql')
    # load registered sqlalchemy datasets
    om = om or om.setup()
    for ds in om.datasets.list(kind=kind, raw=True):
        cnxstr = 'omsql://{ds.name}'.format(**locals())
        conn = om.datasets.get(ds.name, raw=True)
        Connection.connections[cnxstr] = ConnectionShim(cnxstr, conn)
    ipython.magic('sql')



from sklearn.utils import Parallel, delayed


class ParallelStep(object):
    def __init__(self, steps=None, agg=None, n_jobs=-1):
        self.steps = steps
        self.agg = agg or np.mean
        self.n_jobs = n_jobs

    def aggregate(self, values):
        return self.agg(values)

    def __call__(self, value):
        values = Parallel(n_jobs=self.n_jobs)(delayed(pfn)(value) for pfn in self.steps)
        return self.aggregate(values)


class DataPipeline(object):
    def __init__(self, steps):
        self.steps = steps
        self.args = None
        self.kwargs = None

    def set_params(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return self

    def process(self, value=None):
        for stepfn in self.steps:
            value = stepfn(value)
        return value

    def __call__(self):
        return self.process()


